#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from http import HTTPStatus
from typing import List, Generator, Optional, Union

import requests
from requests import Response
from urllib3 import PoolManager, exceptions

from th2_data_services.provider.interfaces.source_api import IHTTPProviderSourceAPI
from th2_data_services.decode_error_handler import UNICODE_REPLACE_HANDLER

logger = logging.getLogger(__name__)


class HTTPProvider5API(IHTTPProviderSourceAPI):
    def __init__(
        self,
        url: str,
        chunk_length: int = 65536,
        decode_error_handler: str = UNICODE_REPLACE_HANDLER,
        char_enc: str = "utf-8",
    ):
        """HTTP Provider5 API.

        Args:
            url: HTTP data source url.
            chunk_length: How much of the content to read in one chunk.
            char_enc: Encoding for the byte stream.
            decode_error_handler: Registered decode error handler.
        """
        self._url = url
        self._char_enc = char_enc
        self._chunk_length = chunk_length
        self._decode_error_handler = decode_error_handler

    def get_url_message_streams(self) -> str:
        """REST-API `messageStreams` call returns a list of message stream names."""
        return f"{self._url}/messageStreams"

    def get_url_find_event_by_id(self, event_id: str) -> str:
        """REST-API `event` call returns a single event with the specified id."""
        return f"{self._url}/event/{event_id}"

    def get_url_find_events_by_id(self, *ids) -> str:
        """REST-API `events` call returns a list of events with the specified ids.

        Note, at a time you can request no more eventSearchChunkSize.

        Deprecated, use `get_url_find_event_by_id` instead.
        """
        query = ""
        for id in ids:
            query += f"ids={id}&"
        return f"{self._url}/events/?{query[:-1]}"

    def get_url_find_message_by_id(self, message_id: str) -> str:
        """REST-API `message` call returns a single message with the specified id."""
        return f"{self._url}/message/{message_id}"

    def get_url_messages_filters(self) -> str:
        """SSE-API `filters/sse-messages` call returns all names of sse message filters.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return f"{self._url}/filters/sse-messages"

    def get_url_events_filters(self) -> str:
        """SSE-API `/filters/sse-events` call returns all names of sse event filters.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return f"{self._url}/filters/sse-events"

    def get_url_message_filter_info(self, filter_name: str) -> str:
        """SSE-API `filters/sse-messages/{filter name}` call returns filter info.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return f"{self._url}/filters/sse-messages/{filter_name}"

    def get_url_event_filter_info(self, filter_name: str) -> str:
        """SSE-API `filters/sse-events/{filter_name}` call returns filter info.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return f"{self._url}/filters/sse-events/{filter_name}"

    def get_url_match_event_by_id(self, event_id: str, filters: str = "") -> str:
        """REST-API `match/event/{id}` call returns boolean value.

        Checks that event with the specified id is matched by filter.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return f"{self._url}/match/event/{event_id}{filters}"

    def get_url_match_message_by_id(self, message_id: str, filters: str = "") -> str:
        """REST-API `match/message/{id}` call returns boolean value.

        Checks that message with the specified id is matched by filter.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return f"{self._url}/match/message/{message_id}{filters}"

    def get_url_search_sse_events(
        self,
        start_timestamp: int,
        end_timestamp: Optional[int] = None,
        parent_event: Optional[str] = None,
        resume_from_id: Optional[str] = None,
        search_direction: Optional[str] = "next",
        result_count_limit: Union[int, float] = None,
        keep_open: Optional[bool] = False,
        limit_for_parent: Union[int, float] = None,
        metadata_only: Optional[bool] = True,
        attached_messages: Optional[bool] = False,
        filters: Optional[str] = None,
    ) -> str:
        """REST-API `search/sse/events` call create a sse channel of event metadata that matches the filter.

        https://github.com/th2-net/th2-rpt-data-provider#sse-requests-api
        """
        kwargs = {
            "startTimestamp": start_timestamp,
            "endTimestamp": end_timestamp,
            "resumeFromId": resume_from_id,
            "parentEvent": parent_event,
            "searchDirection": search_direction,
            "resultCountLimit": result_count_limit,
            "limitForParent": limit_for_parent,
            "keepOpen": keep_open,
            "metadataOnly": metadata_only,
            "attachedMessages": attached_messages,
        }

        query = ""
        url = f"{self._url}/search/sse/events/?"
        for k, v in kwargs.items():
            if v is None:
                continue
            else:
                query += f"&{k}={v}"
        url = f"{url}{query[1:]}"
        if filters is not None:
            url += filters
        return url

    def get_url_search_sse_messages(
        self,
        start_timestamp: int,
        stream: List[str],
        end_timestamp: Optional[int] = None,
        resume_from_id: Optional[str] = None,
        search_direction: Optional[str] = "next",
        result_count_limit: Union[int, float] = None,
        keep_open: bool = False,
        message_id: Optional[List[str]] = None,
        attached_events: bool = False,
        lookup_limit_days: Union[int, float] = None,
        filters: Optional[str] = None,
    ) -> str:
        """REST-API `search/sse/messages` call create a sse channel of messages that matches the filter.

        https://github.com/th2-net/th2-rpt-data-provider#sse-requests-api
        """
        kwargs = {
            "startTimestamp": start_timestamp,
            "endTimestamp": end_timestamp,
            "resumeFromId": resume_from_id,
            "stream": stream,
            "searchDirection": search_direction,
            "resultCountLimit": result_count_limit,
            "keepOpen": keep_open,
            "messageId": message_id,
            "attachedEvents": attached_events,
            "lookupLimitDays": lookup_limit_days,
        }

        query = ""
        url = f"{self._url}/search/sse/messages/?"
        for k, v in kwargs.items():
            if v is None:
                continue
            if k == "stream":
                for s in stream:
                    query += f"&{k}={s}"
            else:
                query += f"&{k}={v}"
        url = f"{url}{query[1:]}"
        if filters is not None:
            url += filters
        return url

    def execute_sse_request(self, url: str) -> Generator[bytes, None, None]:
        """Create stream connection.

        Args:
            url: Url.

        Yields:
             str: Response stream data.
        """
        headers = {"Accept": "text/event-stream"}
        http = PoolManager()
        response = http.request(method="GET", url=url, headers=headers, preload_content=False)

        if response.status != HTTPStatus.OK:
            for s in HTTPStatus:
                if s == response.status:
                    raise exceptions.HTTPError(f"{s.value} {s.phrase} ({s.description})")
            raise exceptions.HTTPError(f"Http returned bad status: {response.status}")

        for chunk in response.stream(self._chunk_length):
            yield chunk

        response.release_conn()

    def execute_request(self, url: str) -> Response:
        """Sends a GET request to provider.

        Args:
            url: Url for a get request to rpt-data-provider.

        Returns:
            requests.Response: Response data.
        """
        return requests.get(url)



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from functools import partial
from typing import List, Iterable, Generator

from grpc._channel import _InactiveRpcError
from th2_grpc_data_provider.data_provider_template_pb2 import (
    EventData,
    MessageData,
)

from th2_data_services import Filter, Data
from th2_data_services.provider.command import ProviderAdaptableCommand
from th2_data_services.provider.exceptions import EventNotFound, MessageNotFound
from th2_data_services.provider.v5.adapters.basic_adapters import GRPCObjectToDictAdapter
from th2_data_services.provider.v5.adapters.event_adapters import DeleteEventWrappersAdapter
from th2_data_services.provider.v5.adapters.message_adapters import DeleteMessageWrappersAdapter
from th2_data_services.provider.v5.interfaces.command import IGRPCProvider5Command

from th2_data_services.provider.v5.data_source.grpc import GRPCProvider5DataSource
from th2_data_services.provider.v5.provider_api import GRPCProvider5API

import logging

logger = logging.getLogger(__name__)


class GetEventByIdGRPCObject(IGRPCProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It retrieves the event by id as GRPC object.

    Returns:
        EventData: Th2 event.
    """

    def __init__(self, id: str):
        """GetEventByIdGRPCObject constructor.

        Args:
            id: Event id.

        """
        super().__init__()
        self._id = id

    def handle(self, data_source: GRPCProvider5DataSource) -> EventData:  # noqa: D102
        api: GRPCProvider5API = data_source.source_api
        event = api.get_event(self._id)

        event = self._handle_adapters(event)
        return event


class GetEventById(IGRPCProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It retrieves the event by id.

    Returns:
        dict: Th2 event.

    Raises:
        EventNotFound: If event by Id wasn't found.
    """

    def __init__(self, id: str, use_stub=False):
        """GetEventById constructor.

        Args:
            id: Event id.
            use_stub: If True the command returns stub instead of exception.

        """
        super().__init__()
        self._id = id
        self._grpc_decoder = GRPCObjectToDictAdapter()
        self._wrapper_deleter = DeleteEventWrappersAdapter()
        self._stub_status = use_stub

    def handle(self, data_source: GRPCProvider5DataSource) -> dict:  # noqa: D102
        try:
            event = GetEventByIdGRPCObject(self._id).handle(data_source)
            event = self._grpc_decoder.handle(event)
            event = self._wrapper_deleter.handle(event)
        except _InactiveRpcError:
            if self._stub_status:
                event = data_source.event_stub_builder.build({data_source.event_struct.EVENT_ID: self._id})
            else:
                logger.error(f"Unable to find the event. Id: {self._id}")
                raise EventNotFound(self._id)

        event = self._handle_adapters(event)
        return event


class GetEventsById(IGRPCProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It retrieves the events by id.

    Returns:
        List[dict]: Th2 events.

    Raises:
        EventNotFound: If any event by Id wasn't found.
    """

    def __init__(self, ids: List[str], use_stub=False):
        """GetEventsById constructor.

        Args:
            ids: Events ids.
            use_stub: If True the command returns stub instead of exception.

        """
        super().__init__()
        self.ids = ids
        self._stub_status = use_stub

    def handle(self, data_source: GRPCProvider5DataSource) -> List[dict]:  # noqa: D102
        response = []
        for event_id in self.ids:
            event = GetEventById(event_id, use_stub=self._stub_status).handle(data_source=data_source)
            event = self._handle_adapters(event)
            response.append(event)

        return response


class GetEventsGRPCObjects(IGRPCProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches events stream as GRPC object by options.

    Returns:
        Iterable[EventData]: Stream of Th2 events.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        end_timestamp: datetime = None,
        parent_event: str = None,
        search_direction: str = "NEXT",
        resume_from_id: str = None,
        result_count_limit: int = None,
        keep_open: bool = False,
        limit_for_parent: int = None,
        attached_messages: bool = False,
        filters: List[Filter] = None,
    ):
        """GetEventsGRPCObjects constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            resume_from_id: Event id from which search starts.
            parent_event: Match events to the specified parent.
            search_direction: Search direction.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
                the one closest to the specified timestamp.
            limit_for_parent: How many children events for each parent do we want to request.
            attached_messages: Gets messages ids which linked to events.
            filters: Filters using in search for messages.

        """
        super().__init__()
        self._start_timestamp = start_timestamp
        self._end_timestamp = end_timestamp
        self._parent_event = parent_event
        self._search_direction = search_direction
        self._resume_from_id = resume_from_id
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._limit_for_parent = limit_for_parent
        self._metadata_only = False
        self._attached_messages = attached_messages
        self._filters = filters

    def handle(self, data_source: GRPCProvider5DataSource) -> Iterable[EventData]:  # noqa: D102
        api: GRPCProvider5API = data_source.source_api

        start_timestamp = int(self._start_timestamp.timestamp() * 10 ** 9)
        end_timestamp = int(self._end_timestamp.timestamp() * 10 ** 9)

        stream_response = api.search_events(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            parent_event=self._parent_event,
            search_direction=self._search_direction,
            resume_from_id=self._resume_from_id,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            limit_for_parent=self._limit_for_parent,
            metadata_only=self._metadata_only,
            attached_messages=self._attached_messages,
            filters=self._filters,
        )
        for response in stream_response:
            if response.WhichOneof("data") == "event":
                response = self._handle_adapters(response)
                yield response.event


class GetEvents(IGRPCProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches events stream by options.

    Returns:
        Iterable[dict]: Stream of Th2 events.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        end_timestamp: datetime = None,
        parent_event: str = None,
        search_direction: str = "NEXT",
        resume_from_id: str = None,
        result_count_limit: int = None,
        keep_open: bool = False,
        limit_for_parent: int = None,
        attached_messages: bool = False,
        filters: List[Filter] = None,
        cache: bool = False,
    ):
        """GetEvents constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            resume_from_id: Event id from which search starts.
            parent_event: Match events to the specified parent.
            search_direction: Search direction.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
                the one closest to the specified timestamp.
            limit_for_parent: How many children events for each parent do we want to request.
            attached_messages: Gets messages ids which linked to events.
            filters: Filters using in search for messages.
            cache: If True, all requested data from rpt-data-provider will be saved to cache.

        """
        super().__init__()
        self._start_timestamp = start_timestamp
        self._end_timestamp = end_timestamp
        self._parent_event = parent_event
        self._search_direction = search_direction
        self._resume_from_id = resume_from_id
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._limit_for_parent = limit_for_parent
        self._metadata_only = False
        self._attached_messages = attached_messages
        self._filters = filters
        self._cache = cache

        self._grpc_decoder = GRPCObjectToDictAdapter()
        self._wrapper_deleter = DeleteEventWrappersAdapter()

    def handle(self, data_source: GRPCProvider5DataSource) -> Data:  # noqa: D102
        source = partial(self.__handle_stream, data_source)
        return Data(source, cache=self._cache)

    def __handle_stream(self, data_source: GRPCProvider5DataSource) -> Generator[dict, None, None]:
        stream = GetEventsGRPCObjects(
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            parent_event=self._parent_event,
            search_direction=self._search_direction,
            resume_from_id=self._resume_from_id,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            limit_for_parent=self._limit_for_parent,
            attached_messages=self._attached_messages,
            filters=self._filters,
        ).handle(data_source)
        for event in stream:
            event = self._grpc_decoder.handle(event)
            event = self._wrapper_deleter.handle(event)
            event = self._handle_adapters(event)
            yield event


class GetMessageByIdGRPCObject(IGRPCProvider5Command, ProviderAdaptableCommand):  # noqa: D102
    """A Class-Command for request to rpt-data-provider.

    It retrieves the message by id as GRPC Object.

    Returns:
        MessageData: Th2 message.
    """

    def __init__(self, id: str):
        """GetMessageByIdGRPCObject constructor.

        Args:
            id: Message id.

        """
        super().__init__()
        self._id = id

    def handle(self, data_source: GRPCProvider5DataSource) -> MessageData:
        api: GRPCProvider5API = data_source.source_api
        response = api.get_message(self._id)
        response = self._handle_adapters(response)
        return response


class GetMessageById(IGRPCProvider5Command, ProviderAdaptableCommand):  # noqa: D102
    """A Class-Command for request to rpt-data-provider.

    It retrieves the message by id.

    Returns:
        dict: Th2 message.

    Raises:
        MessageNotFound: If message by id wasn't found.
    """

    def __init__(self, id: str, use_stub=False):
        """GetMessageById constructor.

        Args:
            id: Message id.
            use_stub: If True the command returns stub instead of exception.

        """
        super().__init__()
        self._id = id
        self._decoder = GRPCObjectToDictAdapter()
        self._wrapper_deleter = DeleteMessageWrappersAdapter()
        self._stub_status = use_stub

    def handle(self, data_source: GRPCProvider5DataSource) -> dict:  # noqa: D102
        try:
            message = GetMessageByIdGRPCObject(self._id).handle(data_source)
            message = self._decoder.handle(message)
            message = self._wrapper_deleter.handle(message)
        except _InactiveRpcError:
            if self._stub_status:
                message = data_source.message_stub_builder.build({data_source.message_struct.MESSAGE_ID: self._id})
            else:
                logger.error(f"Unable to find the message. Id: {self._id}")
                raise MessageNotFound(self._id)
        message = self._handle_adapters(message)
        return message


class GetMessagesById(IGRPCProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It retrieves the messages by id.

    Returns:
        List[dict]: Th2 messages.

    Raises:
        MessageNotFound: If any message by id wasn't found.
    """

    def __init__(self, ids: List[str], use_stub=False):
        """GetMessagesById constructor.

        Args:
            ids: Messages id.
            use_stub: If True the command returns stub instead of exception.

        """
        super().__init__()
        self._ids = ids
        self._stub_status = use_stub

    def handle(self, data_source: GRPCProvider5DataSource) -> List[dict]:  # noqa: D102
        response = []
        for id_ in self._ids:
            message = GetMessageById(id_, use_stub=self._stub_status).handle(data_source)
            message = self._handle_adapters(message)
            response.append(message)

        return response


class GetMessagesGRPCObject(IGRPCProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches messages stream as GRPC object by options.

    Returns:
        Iterable[MessageData]: Stream of Th2 messages.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        stream: List[str],
        end_timestamp: datetime = None,
        resume_from_id: str = None,
        search_direction: str = "NEXT",
        result_count_limit: int = None,
        keep_open: bool = False,
        filters: List[Filter] = None,
    ):
        """GetMessagesGRPCObject constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            stream: Alias of messages.
            resume_from_id: Message id from which search starts.
            search_direction: Search direction.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
            filters: Filters using in search for messages.

        """
        super().__init__()
        self._start_timestamp = start_timestamp
        self._end_timestamp = end_timestamp
        self._stream = stream
        self._resume_from_id = resume_from_id
        self._search_direction = search_direction
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._filters = filters

    def handle(self, data_source: GRPCProvider5DataSource) -> List[MessageData]:
        api = data_source.source_api

        start_timestamp = int(self._start_timestamp.timestamp() * 10 ** 9)
        end_timestamp = int(self._end_timestamp.timestamp() * 10 ** 9)

        stream_response = api.search_messages(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            stream=self._stream,
            resume_from_id=self._resume_from_id,
            search_direction=self._search_direction,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            filters=self._filters,
        )
        for response in stream_response:
            if response.WhichOneof("data") == "message":
                response = self._handle_adapters(response)
                yield response.message


class GetMessages(IGRPCProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches messages stream by options.

    Returns:
        Iterable[dict]: Stream of Th2 messages.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        stream: List[str],
        end_timestamp: datetime = None,
        resume_from_id: str = None,
        search_direction: str = "NEXT",
        result_count_limit: int = None,
        keep_open: bool = False,
        filters: List[Filter] = None,
        cache: bool = False,
    ):
        """GetMessages constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            stream: Alias of messages.
            resume_from_id: Message id from which search starts.
            search_direction: Search direction.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
            filters: Filters using in search for messages.
            cache: If True, all requested data from rpt-data-provider will be saved to cache.

        """
        super().__init__()
        self._start_timestamp = start_timestamp
        self._end_timestamp = end_timestamp
        self._stream = stream
        self._resume_from_id = resume_from_id
        self._search_direction = search_direction
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._filters = filters
        self._cache = cache

        self._decoder = GRPCObjectToDictAdapter()
        self._wrapper_deleter = DeleteMessageWrappersAdapter()

    def handle(self, data_source: GRPCProvider5DataSource) -> Data:
        source = partial(self.__handle_stream, data_source)
        return Data(source, cache=self._cache)

    def __handle_stream(self, data_source: GRPCProvider5DataSource) -> Iterable[dict]:
        stream = GetMessagesGRPCObject(
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            stream=self._stream,
            resume_from_id=self._resume_from_id,
            search_direction=self._search_direction,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            filters=self._filters,
        ).handle(data_source)
        for message in stream:
            message = self._decoder.handle(message)
            message = self._wrapper_deleter.handle(message)
            message = self._handle_adapters(message)
            yield message



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pprint
from _warnings import warn
from typing import Union, List

from th2_data_services.interfaces.adapter import IMessageAdapter
from th2_data_services.provider.v5.struct import Provider5MessageStruct, provider5_message_struct


class DeleteMessageWrappersAdapter(IMessageAdapter):
    """Adapter that deletes unnecessary wrappers in messages.

    It used for the message to which an AdaptorGRPCObjectToDict has been applied.
    """

    def __init__(self, message_struct: Provider5MessageStruct = provider5_message_struct):
        """AdapterDeleteMessageWrappers constructor.

        Args:
            message_struct: Message struct.
        """
        self._message_struct = message_struct

    def handle(self, message: dict) -> dict:
        """Deletes unnecessary wrappers for field message_id.

        Args:
            message: Message.

        Returns:
            Message without wrappers.
        """
        message_id_field = self._message_struct.MESSAGE_ID

        message_id = message[message_id_field]

        session = message_id[self._message_struct.CONNECTION_ID][self._message_struct.SESSION_ALIAS]
        direction = message_id[self._message_struct.DIRECTION]
        sequence = message_id[self._message_struct.SEQUENCE]

        message_id = f"{session}:{direction}:{sequence}"
        message[message_id_field] = message_id

        return message


class CodecPipelinesAdapter(IMessageAdapter):
    """Adapter for codec-pipeline messages from provider v5.

    Codec-pipeline messages have sub-messages in the body.
    This adapter used for split codec-pipeline message to separate messages.
    """

    def __init__(self, ignore_errors=False):
        """AdapterCodecPipelines constructor.

        Args:
            ignore_errors: If True it will ignore errors and return message as is.
        """
        self._ignore_errors = ignore_errors

    def handle(self, message: dict) -> Union[List[dict], dict]:
        """Adapter handler.

        Args:
            message: Th2Message dict.

        Returns:
            Th2Message dict.
        """
        msg_type = message.get("messageType")
        if msg_type is None:
            if self._ignore_errors:
                warn(
                    "Please note, some messages don't have a messageType field. Perhaps a codec didn't decode them.",
                    stacklevel=3,
                )
                return message
            else:
                raise ValueError(
                    "The messages doesn't have a messageType field. Message:\n" f"{pprint.pformat(message)}"
                )

        if "/" not in msg_type:
            return message

        body = message["body"]
        if not body:
            return message

        sub_messages = []
        fields = body["fields"]
        if not fields:
            return message

        for sub_msg in fields:
            split_msg_name = sub_msg.split("-")
            if len(split_msg_name) > 1:
                sub_msg_type, index = "".join(split_msg_name[:-1]), int(split_msg_name[-1])
            else:
                index = msg_type.split("/").index(sub_msg) + 1
                sub_msg_type = sub_msg

            new_record = message.copy()

            metadata = new_record["body"]["metadata"].copy()
            id_field = metadata["id"].copy()
            id_field["subsequence"] = [index]
            metadata["id"] = id_field

            body_fields = fields[sub_msg]
            metadata.update(body_fields.get("metadata", {}))

            body = {"metadata": metadata}
            if body_fields.get("messageValue"):
                body = {**body_fields["messageValue"], **body}
            elif body_fields.get("fields"):
                body = {**body_fields["fields"], **body}
            else:
                body = {"fields": {}, **body}

            new_record["body"] = body
            new_record["body"]["metadata"]["messageType"] = sub_msg_type
            new_record["messageType"] = sub_msg_type
            new_record["messageId"] = f"{message['messageId']}.{index}"
            sub_messages.append(new_record)

        return sub_messages



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sseclient
from sseclient import _FIELD_SEPARATOR


class SSEClient(sseclient.SSEClient):
    """Patch for sseclient-py to get availability to configure decode error handler."""

    def __init__(self, event_source, char_enc="utf-8", decode_errors_handler="strict"):
        """Initialize the SSE client over an existing, ready to consume event source.

        The event source is expected to be a binary stream and have a close()
        method. That would usually be something that implements
        io.BinaryIOBase, like an httplib or urllib3 HTTPResponse object.
        """
        super(SSEClient, self).__init__(event_source, char_enc)
        self._decode_errors_handler = decode_errors_handler

    def events(self):
        """Returns events in generator style."""
        for chunk in self._read():
            event = sseclient.Event()
            # Split before decoding so splitlines() only uses \r and \n
            for line in chunk.splitlines():
                # Decode the line.
                line = line.decode(self._char_enc, self._decode_errors_handler)

                # Lines starting with a separator are comments and are to be
                # ignored.
                if not line.strip() or line.startswith(_FIELD_SEPARATOR):
                    continue

                data = line.split(_FIELD_SEPARATOR, 1)
                field = data[0]

                # Ignore unknown fields.
                if field not in event.__dict__:
                    self._logger.debug("Saw invalid field %s while parsing " "Server Side Event", field)
                    continue

                if len(data) > 1:
                    # From the spec:
                    # "If value starts with a single U+0020 SPACE character,
                    # remove it from value."
                    if data[1].startswith(" "):
                        value = data[1][1:]
                    else:
                        value = data[1]
                else:
                    # If no value is present after the separator,
                    # assume an empty value.
                    value = ""

                # The data field may come over multiple lines and their values
                # are concatenated with each other.
                if field == "data":
                    event.__dict__[field] += value + "\n"
                else:
                    event.__dict__[field] = value

            # Events with no data are not dispatched.
            if not event.data:
                continue

            # If the data field ends with a newline, remove it.
            if event.data.endswith("\n"):
                event.data = event.data[0:-1]

            # Empty event names default to 'message'
            event.event = event.event or "message"

            # Dispatch the event
            self._logger.debug("Dispatching %s...", event)
            yield event



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Dict, Optional

from treelib import Node, Tree

from th2_data_services import Data
from th2_data_services.events_tree import EventsTree
from th2_data_services.interfaces.events_tree.events_tree_collection import EventsTreeCollection
from th2_data_services.provider.interfaces.data_source import IProviderDataSource


class ParentEventsTreeCollection(EventsTreeCollection):
    """ParentEventsTreeCollections is a class like an EventsTreeCollections.

    ParentEventsTree contains all parent events that are referenced.
    """

    def __init__(
        self, data: Data, data_source: IProviderDataSource = None, preserve_body: bool = False, stub: bool = False
    ):
        """ParentEventsTreeCollection constructor.

        Args:
            data: Data object.
            data_source: Data Source object.
            preserve_body: If True then save body of event.
            stub: If True it will create stub when event is broken.
        """
        super().__init__(data, data_source, preserve_body, stub)

    def _build_trees(self, nodes: Dict[Optional[str], List[Node]]) -> None:
        """Builds trees and saves detached events.

        Args:
            nodes: Events nodes.
        """
        roots = []
        for node in nodes[None]:  # None - is parent_event_id for root events.
            if nodes[node.identifier]:
                tree = Tree()
                tree.add_node(node)
                roots.append(tree)
                event_id: str = node.identifier
                self._fill_tree(nodes, tree, event_id)
        nodes.pop(None)

        self._roots = [EventsTree(root) for root in roots]
        self._detached_nodes = nodes

    def _fill_tree(self, nodes: Dict[Optional[str], List[Node]], current_tree: Tree, parent_id: str) -> None:
        """Fills tree recursively.

        Args:
            nodes: Events nodes.
            current_tree: Tree for fill.
            parent_id: Parent even id.
        """
        for node in nodes[parent_id]:
            event_id: str = node.identifier
            if nodes.get(event_id):
                current_tree.add_node(node, parent=parent_id)
                self._fill_tree(nodes, current_tree, event_id)  # recursive fill
        nodes.pop(parent_id)



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from collections import defaultdict
from typing import List, Dict, Optional, Union, Generator, Tuple, Callable

from treelib import Node, Tree

from th2_data_services import Data
from th2_data_services.events_tree.events_tree import EventsTree
from th2_data_services.events_tree.events_tree import Th2Event
from th2_data_services.events_tree.exceptions import EventIdNotInTree
from th2_data_services.provider.interfaces.data_source import IProviderDataSource
from th2_data_services.provider.v5.command_resolver import resolver_get_events_by_id


class EventsTreeCollection(ABC):
    """EventsTreeCollection objective is building 'EventsTree's and storing them.

    EventsTreeCollection stores all EventsTree. You can to refer to each of them.
    """

    def __init__(
        self, data: Data, data_source: IProviderDataSource = None, preserve_body: bool = False, stub: bool = False
    ):
        """EventsTreeCollection constructor.

        Args:
            data: Data object.
            data_source: Data Source object.
            preserve_body: If True it will preserve 'body' field in the Events.
            stub: If True it will create stub when event is broken.
        """
        self._preserve_body = preserve_body
        self._roots: List[EventsTree] = []
        self._parentless: Optional[List[EventsTree]] = None
        self._detached_nodes: Dict[Optional[str], List[Node]] = defaultdict(list)  # {parent_event_id: [Node1, ..]}
        self._stub_status = stub
        self._data_source = data_source

        events_nodes = self._build_event_nodes(data)
        self._build_trees(events_nodes)

        if data_source is not None:
            self._recover_unknown_events()

    def get_parentless_trees(self) -> List[EventsTree]:
        """Builds and returns parentless trees by detached events.

        Detached events will be removed from the tree.

        Returns:
            List of parentless trees if they exist, otherwise empty list.
        """
        if self._parentless is not None:
            return self._parentless
        else:
            self._parentless = self._build_parentless_trees()
            return self._parentless

    def _build_parentless_trees(self) -> List[EventsTree]:
        """Builds parentless trees by detached events.

        Returns:
            Parentless trees.
        """
        self._parentless = []

        stub_roots = list(self._detached_nodes.keys())
        for nodes in self._detached_nodes.values():
            for node in nodes:
                if node.identifier in stub_roots:
                    stub_roots.remove(node.identifier)

        for id_ in stub_roots:
            tree = Tree()
            tree.create_node(tag="Stub", identifier=id_, data={})
            self._fill_tree(self._detached_nodes, tree, id_)
            self._parentless.append(EventsTree(tree))

        return self._parentless

    def _build_event_nodes(self, data: Data) -> Dict[Optional[str], List[Node]]:
        """Builds event nodes and group them by parent_event_id.

        Args:
            data: Data.

        Returns:
            Nodes.
        """
        events_nodes: Dict[Optional[str], List[Node]] = defaultdict(list)  # {parent_event_id: [Node1, Node2, ..]}

        for event in self._parse_events(data):
            parent_event_id: str = self._get_parent_event_id(event)
            node = self._build_node(event)
            events_nodes[parent_event_id].append(node)

        return events_nodes

    def _build_trees(self, nodes: Dict[Optional[str], List[Node]]) -> None:
        """Builds trees and saves detached events.

        Args:
            nodes: Events nodes.
        """
        roots = []
        for root_node in nodes[None]:  # None - is parent_event_id for root events.
            tree = Tree()
            tree.add_node(root_node)
            roots.append(tree)
            root_event_id: str = root_node.identifier
            self._fill_tree(nodes, tree, root_event_id)
        nodes.pop(None)

        self._roots = [EventsTree(root) for root in roots]
        self._detached_nodes = nodes

    def _fill_tree(self, nodes: Dict[Optional[str], List[Node]], current_tree: Tree, parent_id: str) -> None:
        """Fills tree recursively.

        Args:
            nodes: Events nodes.
            current_tree: Tree for fill.
            parent_id: Parent even id.
        """
        for node in nodes[parent_id].copy():
            event_id: str = node.identifier
            if event_id not in current_tree:
                current_tree.add_node(node, parent=parent_id)
            nodes[parent_id].remove(node)
            self._fill_tree(nodes, current_tree, event_id)  # Recursive fill.
        nodes.pop(parent_id)

    def _parse_event(self, event: Th2Event) -> Th2Event:
        """Parses event.

        Args:
            event: Event.

        Returns:
            Parsed event.
        """
        if not self._preserve_body:
            try:
                event.pop("body")
            except KeyError:
                pass
        return event

    def _parse_events(self, data: Union[Data, List[dict]]) -> Generator[dict, None, None]:
        """Parses events.

        Args:
            data: Data.

        Returns:
            Parsed events iterator.
        """
        for event in data:
            event = self._parse_event(event)
            yield event

    def _build_node(self, event: dict) -> Node:
        """Builds event as dict into event as Node.

        Args:
            event: Event.

        Returns:
            Node.
        """
        event_id: str = self._get_event_id(event)
        event_name: str = self._get_event_name(event)

        node = Node(tag=event_name, identifier=event_id, data=event)
        return node

    def append_event(self, event: dict) -> None:
        """Appends event into tree.

        Args:
            event: Event.
        """
        event: dict = self._parse_event(event)
        node: Node = self._build_node(event)
        parent_event_id: str = self._get_parent_event_id(event)

        if parent_event_id is not None:
            events_trees = list(filter(lambda tree: parent_event_id in tree, self._roots))
            if events_trees:
                event_tree = events_trees[0]
                event_tree._append_node(node, parent_event_id)
                self._fill_tree(self._detached_nodes, event_tree._tree, parent_event_id)
            else:
                self._detached_nodes[parent_event_id].append(node)
        else:
            tree = Tree()
            tree.add_node(node)
            self._roots.append(EventsTree(tree))
            self._fill_tree(self._detached_nodes, tree, node.identifier)

    @property
    def detached_events(self) -> dict:
        """Returns detached events as a dict with a view {'parent_id': ['referenced event', ...]}."""
        return {id_: [node.data for node in nodes] for id_, nodes in self._detached_nodes.items()}

    def get_roots_ids(self) -> List[str]:
        """Returns roots ids."""
        return [tree.get_root_id() for tree in self._roots]

    def get_trees(self) -> List[EventsTree]:
        """Returns the list of trees inside the collection."""
        return self._roots

    def get_root_by_id(self, id) -> EventsTree:
        """Returns a root tree by id as EventsTree class.

        Args:
            id: Root id.

        Returns:
            Root tree.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            if tree.get_root_id() == id:
                return tree
        raise EventIdNotInTree(id)

    def show(self):
        """Prints all EventsTrees as tree view.

        For example:

        ```
        Root
            |___ C01
            |    |___ C11
            |         |___ C111
            |         |___ C112
            |___ C02
            |___ C03
            |    |___ C31
        ```
        """
        trees = self.get_trees()
        for tree in trees:
            tree.show()

    @abstractmethod
    def _get_parent_event_id(self, event):
        """Gets parent event id from the event."""

    @abstractmethod
    def _get_event_id(self, event):
        """Gets event id from the event."""

    @abstractmethod
    def _get_event_name(self, event):
        """Gets event name from the event."""

    def __len__(self) -> int:
        """Returns the number of all events, including detached events."""
        return self.len_trees + self.len_detached_events

    def __contains__(self, event_id: str):
        return any([event_id in tree for tree in self._roots])

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(trees={len(self.get_roots_ids())}, "
            f"events={len(self)}[trees={self.len_trees}, detached={self.len_detached_events}])"
        )

    def summary(self) -> str:
        """Returns the collection summary."""
        return self.__repr__()

    @property
    def len_trees(self) -> int:
        """Returns number of events in the trees inside the collection."""
        return sum([len(root) for root in self._roots])

    @property
    def len_detached_events(self) -> int:
        """Returns number of detached events in the collection."""
        return sum([len(de_lst) for de_lst in self.detached_events.values()])

    def get_all_events_iter(self) -> Generator[Th2Event, None, None]:
        """Returns all events from the trees as iterator."""
        for tree in self._roots:
            yield from tree.get_all_events()

    def get_all_events(self) -> List[Th2Event]:
        """Returns all events from the trees."""
        return [event for tree in self._roots for event in tree.get_all_events()]

    def get_event(self, id: str) -> Optional[Th2Event]:
        """Returns an event by id.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_event(id)
            except EventIdNotInTree:
                continue
        raise EventIdNotInTree(id)

    def get_leaves(self) -> Tuple[Th2Event]:
        """Returns all trees leaves."""
        return tuple(self.get_leaves_iter())

    def get_leaves_iter(self) -> Generator[Th2Event, None, None]:
        """Returns all trees leaves as iterator."""
        for tree in self._roots:
            yield from tree.get_leaves_iter()

    def get_children(self, id: str) -> Tuple[Th2Event]:
        """Returns children for the event by its id.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_children(id)
            except EventIdNotInTree:
                continue
        raise EventIdNotInTree(id)

    def get_children_iter(self, id: str) -> Generator[Th2Event, None, None]:
        """Returns children as iterator for the event by its id.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        is_iter = False
        for tree in self._roots:
            try:
                yield from tree.get_children_iter(id)
                is_iter = True
            except EventIdNotInTree:
                continue
        if not is_iter:
            raise EventIdNotInTree(id)

    def get_parent(self, id: str) -> Th2Event:
        """Returns a parent for the event by its id.

        Args:
            id: Event id.

        Raises:
            NodeIDAbsentError: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_parent(id)
            except EventIdNotInTree:
                continue
        raise EventIdNotInTree(id)

    def get_full_path(self, id: str, field: str = None) -> List[Union[str, Th2Event]]:  # noqa: D412
        """Returns the full path for the event by its id in the right order.

        Examples:

        Imagine we have the following tree.

        ```
        Harry
        ├── Bill
        └── Jane
            ├── Diane
            │   └── Mary
            └── Mark
        ```

        ```
        collection.get_full_path('Jane', id)
        ['Harry-event-id', 'Jane-event-id']

        collection.get_full_path('Jane', name)
        ['Harry-event-name', 'Jane-event-name']

        collection.get_full_path('Jane')
        ['Harry-event', 'Jane-event']
        ```

        Args:
            id: Event id.
            field: Field of event.

        Returns:
            Full path of event.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_full_path(id, field)
            except EventIdNotInTree:
                continue
        raise EventIdNotInTree(id)

    def get_ancestors(self, id: str) -> List[Th2Event]:
        """Returns all event's ancestors in right order.

        Args:
            id: Event id.

        Returns:
            All event's ancestors.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_ancestors(id)
            except EventIdNotInTree:
                continue
        raise EventIdNotInTree(id)

    def find_ancestor(self, id: str, filter: Callable) -> Optional[Th2Event]:
        """Finds the ancestor of an event.

        Args:
            id: Event id.
            filter: Filter function

        Returns:
            Ancestor of Event.
        """
        for tree in self._roots:
            ancestor = tree.find_ancestor(id, filter)
            if ancestor is not None:
                return ancestor
        return None

    def findall_iter(
        self,
        filter: Callable,
        stop: Callable = None,
        max_count: int = None,
    ) -> Generator[Th2Event, None, None]:
        """Searches events matches as iterator.

        - The search uses 'filter' which is a filtering function.
        - Optionally, the search uses 'stop' which is a stopping function.
        If 'stop' function returns 'True' then search is complete.
        - 'max_count' is a parameter that limits the search to a specified count.

        Args:
            filter: Filter function.
            stop: Stop function. If None searches for all nodes in the trees.
            max_count: Max count of matched events. Stops searching when `max_count` will be reached.

        Yields:
            Matching events.
        """
        for tree in self._roots:
            yield from tree.findall_iter(filter=filter, stop=stop, max_count=max_count)

    def findall(
        self,
        filter: Callable,
        stop: Callable = None,
        max_count: int = None,
    ) -> List[Th2Event]:
        """Searches events matches.

        - The search uses 'filter' which is a filtering function.
        - Optionally, the search uses 'stop' which is a stopping function.
        If 'stop' function returns 'True' then search is complete.
        - 'max_count' is a parameter that limits the search to a specified count.

        Args:
            filter: Filter function.
            stop: Stop function. If None searches for all nodes in the trees.
            max_count: Max count of matched events. Stops searching when `max_count` will be reached.

        Returns:
            Matching events.
        """
        result = []
        for tree in self._roots:
            events = tree.findall(filter=filter, stop=stop, max_count=max_count)
            if result is not None:
                result += [*events]
        return result

    def find(self, filter: Callable, stop: Callable = None) -> Optional[Th2Event]:
        """Searches the first event match.

        - The search uses 'filter' which is a filtering function.
        - Optionally, the search uses 'stop' which is a stopping function.
        If 'stop' function returns 'True' then search is complete.

        Args:
            filter: Filter function.
            stop: Stop function. If None searches for all nodes in the trees.

        Returns:
            One matching event.
        """
        for tree in self._roots:
            event = tree.find(filter=filter, stop=stop)
            if event is not None:
                return event
        return None

    def get_subtree(self, id: str) -> "EventsTree":
        """Returns subtree of the event by its id.

        Args:
            id: Event id.

        Returns:
            Subtree.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_subtree(id)
            except EventIdNotInTree:
                continue
        raise EventIdNotInTree(id)

    def _recover_unknown_events(self) -> None:
        """Loads missed events and recover events."""
        instance_command = resolver_get_events_by_id(self._data_source)

        previous_detached_events = list(self.detached_events.keys())
        while previous_detached_events:
            called_command = instance_command(self.detached_events.keys(), use_stub=self._stub_status)
            events = self._data_source.command(called_command)

            for event in events:
                if not self._get_event_name(event) == "Broken_Event":
                    self.append_event(event)

            if previous_detached_events == list(self.detached_events.keys()):
                break
            previous_detached_events = list(self.detached_events.keys())



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from th2_data_services.interfaces.adapter import IEventAdapter
from th2_data_services.provider.v5.struct import Provider5EventStruct, provider5_event_struct


class DeleteEventWrappersAdapter(IEventAdapter):
    """Adapter that deletes unnecessary wrappers in events.

    It used for events to which an AdaptorGRPCObjectToDict has been applied.
    """

    def __init__(self, event_struct: Provider5EventStruct = provider5_event_struct):
        """AdapterDeleteEventWrappers constructor.

        Args:
            event_struct: Event struct.
        """
        self._event_struct = event_struct

    def handle(self, event: dict) -> dict:
        """Deletes unnecessary wrappers for fields eventId, parentEventId and BatchId.

        Args:
            event: Event.

        Returns:
            Event without wrappers.
        """
        event_id_field = self._event_struct.EVENT_ID
        parent_event_id_field = self._event_struct.PARENT_EVENT_ID
        batch_id_field = self._event_struct.BATCH_ID

        event_id = self.__get_id_from_wrapper(event, event_id_field)
        parent_event_id = self.__get_id_from_wrapper(event, parent_event_id_field)
        batch_id = self.__get_id_from_wrapper(event, batch_id_field)

        if event_id:
            event[event_id_field] = event_id
        if parent_event_id:
            event[parent_event_id_field] = parent_event_id
        if batch_id:
            event[batch_id_field] = batch_id

        return event

    @staticmethod
    def __get_id_from_wrapper(event: dict, field: str):
        """Opens the wrapper and getting the id."""
        wrapper: dict = event.get(field)
        if wrapper:
            return wrapper["id"]
        return None



class Th2GUIReport:
    """Class for creating gui link by event ID or message ID."""

    def __init__(self, provider_link: str):
        """Th2GUIReport constructor.

        Args:
            provider_link (str): link to provider.
        """
        self._provider_link = self.__normalize_link(provider_link)

    def __normalize_link(self, link: str) -> str:
        """Bringing links to a single form.

        Add 'http://' to the beginning of the link.
        Add slash to the ending of link.

        Args:
            link (str): link for editing.

        Returns:
            Normalize link.
        """
        find_http = link.startswith("http", 0)
        if find_http is False:
            link = "http://" + link

        if link[-1] != "/":
            link = link + "/"

        return link

    def get_event_link(self, event_id: str) -> str:
        """Creates the link with event id.

        Args:
            event_id (str): id for adding in link.

        Returns:
            GUI link to event.
        """
        return f"{self._provider_link}?eventId={event_id}"

    def get_message_link(self, message_id: str) -> str:
        """Creates the link with message id.

        Args:
            message_id (str): id for adding in link.

        Returns:
            GUI link to message.
        """
        return f"{self._provider_link}?messageId={message_id}"



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Interfaces for Provider Data Source."""

from abc import abstractmethod

import requests
import urllib3

from th2_data_services.provider.interfaces.command import IGRPCProviderCommand, IHTTPProviderCommand, IProviderCommand
from th2_data_services.interfaces.data_source import IDataSource
from th2_data_services.provider.interfaces.source_api import (
    IHTTPProviderSourceAPI,
    IGRPCProviderSourceAPI,
    IProviderSourceAPI,
)
from th2_data_services.provider.interfaces.struct import IEventStruct, IMessageStruct
from th2_data_services.provider.interfaces.stub_builder import IEventStub, IMessageStub


class IProviderDataSource(IDataSource):
    def __init__(
        self,
        url: str,
        event_struct: IEventStruct,
        message_struct: IMessageStruct,
        event_stub_builder: IEventStub,
        message_stub_builder: IMessageStub,
    ):
        """Interface of DataSource that provides work with rpt-data-provider.

        Args:
            url: Url address to data provider.
            event_struct: Event struct class.
            message_struct: Message struct class.
            event_stub_builder: Event stub builder class.
            message_stub_builder: Message stub builder class.
        """
        if url[-1] == "/":
            url = url[:-1]
        self._url = url
        self._event_struct = event_struct
        self._message_struct = message_struct
        self._event_stub_builder = event_stub_builder
        self._message_stub_builder = message_stub_builder

    @property
    def url(self) -> str:
        """str: URL of rpt-data-provider."""
        return self._url

    @property
    def event_struct(self) -> IEventStruct:
        """Returns event structure class."""
        return self._event_struct

    @property
    def message_struct(self) -> IMessageStruct:
        """Returns message structure class."""
        return self._message_struct

    @property
    def event_stub_builder(self) -> IEventStub:
        """Returns event stub template."""
        return self._event_stub_builder

    @property
    def message_stub_builder(self) -> IMessageStub:
        """Returns message stub template."""
        return self._message_stub_builder

    @abstractmethod
    def command(self, cmd: IProviderCommand):
        """Execute the transmitted command."""

    @property
    @abstractmethod
    def source_api(self) -> IProviderSourceAPI:
        """Returns Provider API."""


class IHTTPProviderDataSource(IProviderDataSource):
    """Interface of DataSource that provides work with rpt-data-provider via HTTP."""

    @abstractmethod
    def command(self, cmd: IHTTPProviderCommand):
        """Execute the transmitted HTTP command."""

    def check_connect(self, timeout: (int, float)) -> None:
        """Checks whether url is working.

        Args:
            timeout: How many seconds to wait for the server to send data before giving up.

        Raises:
            urllib3.exceptions.HTTPError: If unable to connect to host.
        """
        try:
            requests.get(self.url, timeout=timeout)
        except ConnectionError as error:
            raise urllib3.exceptions.HTTPError(f"Unable to connect to host '{self.url}'\nReason: {error}")

    @property
    @abstractmethod
    def source_api(self) -> IHTTPProviderSourceAPI:
        """Returns HTTP Provider API."""


class IGRPCProviderDataSource(IProviderDataSource):
    """Interface of DataSource that provides work with rpt-data-provider via GRPC."""

    @abstractmethod
    def command(self, cmd: IGRPCProviderCommand):
        """Execute the transmitted GRPC command."""

    @property
    @abstractmethod
    def source_api(self) -> IGRPCProviderSourceAPI:
        """Returns GRPC Provider API."""



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import copy
import pickle
import pprint
from os import rename
from pathlib import Path
from time import time
from typing import Callable, Dict, Generator, List, Optional, Union, Iterable, Any
from weakref import finalize
import logging

logger = logging.getLogger(__name__)

DataSet = Union["Data", Iterable, Callable[..., Generator[dict, None, None]]]
WorkFlow = List[Dict[str, Union[Callable, str]]]
DataGenerator = Generator[Any, None, None]


class _DataLogger(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        return "Data[%s] %s" % (self.extra["id"], msg), kwargs


class Data:
    """A wrapper for data/data_stream.

    The class provides methods for working with data as a stream.

    Such approach to data analysis called streaming transformation.
    """

    def __init__(self, data: DataSet, cache: bool = False, workflow: WorkFlow = None):
        """Data constructor.

        Args:
            data: Data source. Any iterable, Data object or function that creates generator.
            cache: Set True if you want to write and read from cache.
            workflow: Workflow.
        """
        self._id = id(self)
        self._cache_filename = f"{self._id}:{time()}.pickle"
        self._cache_path = Path(f"./temp/{self._cache_filename}").resolve()
        self._len = None
        self._data_stream = data
        self._workflow = [] if workflow is None else workflow  # Normally it has empty list or one Step.
        self._length_hint = None  # The value is populated when we use limit method.
        self._cache_status = cache
        self._finalizer = finalize(self, self.__remove)
        self._logger = _DataLogger(logger, {"id": self._id})
        # It used to indicate the number of current iteration of the Data object.
        # It's required if the same instance iterates several times in for-in loops.
        self.iter_num = 0
        self.stop_iteration = None

        self._logger.info(
            "New data object with data stream = '%s', cache = '%s' initialized", id(self._data_stream), cache
        )

    def __remove(self):
        """Data class destructor."""
        if self.__is_cache_file_exists():
            self.__delete_cache()
        del self._data_stream

    def __delete_cache(self) -> None:
        """Removes cache file."""
        path = self.get_cache_filepath()
        if path.exists():
            self._logger.debug("Deleting cache file '%s'" % path)
            path.unlink()

    def __delete_pending_cache(self) -> None:
        """Removes cache file."""
        path = self.get_pending_cache_filepath()
        if path.exists():
            self._logger.debug("Deleting cache file '%s'" % path)
            path.unlink()

    @property
    def len(self) -> int:
        """int: How many records in the Data stream.

        Notes:
        1. It is a wasteful operation if you are performing it on the Data object that has never been iterated before.

        2. If you want just to check emptiness, use is_empty property instead.
        """
        return self._len if self._len is not None else self.__calc_len()

    @property
    def is_empty(self) -> bool:
        """bool: Indicates that the Data object doesn't contain data."""
        for _ in self.__load_data():
            return False
        return True

    def __calc_len(self) -> int:
        for _ in self:
            pass
        return self._len

    def __length_hint__(self):
        """This hint used by list() to calculate the size of the bytes for the list."""
        if self._len is not None:
            return self._len
        elif self._length_hint is not None:
            return self._length_hint
        else:
            # 2**13, though 8 - is a default value in CPython.
            # We usually have large number of data.
            return 8192

    def _iter_logic(self):
        interruption = True
        if self._cache_status and self.__is_cache_file_exists():
            is_data_writes_cache = False
        else:
            is_data_writes_cache = True

        try:
            for record in self.__load_data(self._cache_status):
                yield record
            else:
                # Loop successfully finished. Do not delete cache file.
                self._logger.debug("Successfully iterated")
                interruption = False

        finally:
            if interruption:

                if self.stop_iteration:  # When limit was reached.
                    # You can save _len in this case because iteration was stopped by limit.
                    self._logger.info("Iteration was interrupted because limit reached")

                else:  # When something went wrong but NOT StopIteration
                    self._logger.info("Iteration was interrupted")
                    # You shouldn't save _len in this case because iteration was interrupted.
                    if self.iter_num == 1:
                        self._len = None

                # Delete cache if it was interrupted and the file was not complete.
                # https://exactpro.atlassian.net/browse/TH2-3546
                if is_data_writes_cache:
                    self._logger.info("The cache file is not written to the end. Delete tmp cache file")
                    self.__delete_pending_cache()
                else:  # Data reads cache.
                    from th2_data_services import INTERACTIVE_MODE  # To escape circular import problem.

                    # Do not delete cache file if it's interactive mode and Data has read cache.
                    if not INTERACTIVE_MODE:
                        self.__delete_cache()

            self.iter_num -= 1
            self.stop_iteration = False

    def __iter__(self) -> DataGenerator:
        self.stop_iteration = False
        self.iter_num += 1
        self._logger.info("Starting iteration, iter_num = %s", self.iter_num)
        if self._len is None and self.iter_num == 1:
            self._len = 0
            for record in self._iter_logic():
                self._len += 1
                yield record
        else:
            # Do not calculate self._len if it is not None.
            yield from self._iter_logic()

    def _build_workflow(self, workflow):
        """Updates limit callbacks each time when Data object is iterated.

        It used to have possibility iterate the same Data object several times in the loops.
        """
        new_workflow = copy.deepcopy(workflow)
        for w in new_workflow[::-1]:
            if w["type"] == "limit":
                w["callback"] = self._build_limit_callback(w["callback"].limit)

        return new_workflow

    def __load_data(self, cache: bool = False) -> DataGenerator:
        """Loads data from cache or data.

        Args:
            cache: Flag if you what to write and read from cache.

        Returns:
            obj: Generator
        """
        if cache and self.__is_cache_file_exists():
            self._logger.info("Iterating using own cache file '%s'" % self.get_cache_filepath())
            data_stream = self.__load_file(self.get_cache_filepath())
            yield from data_stream
        else:
            data_stream = self._data_stream() if callable(self._data_stream) else self._data_stream
            workflow = self._build_workflow(self._workflow)

            if self.__check_file_recording():
                # Do not read from the cache file if it has PENDING status (if the file is not filled yet).
                # It used to handle case when Data object iterates in the loop several times.
                cache = False

            yield from self.__change_data(data_stream=data_stream, workflow=workflow, cache=cache)

    def __check_file_recording(self) -> bool:
        """Checks whether there is a current recording in the file.

        Returns:
            bool: File recording status.
        """
        path = self.get_pending_cache_filepath()
        return path.is_file()

    def get_pending_cache_filepath(self) -> Path:
        """Returns filepath for a pending cache file."""
        filepath = self.get_cache_filepath()
        return filepath.with_name("[PENDING]" + self._cache_filename)

    def get_cache_filepath(self) -> Optional[Path]:
        """Returns filepath for a cache file."""
        return self._cache_path

    def _iterate_modified_data_stream(self, data_stream: DataGenerator, workflow: WorkFlow) -> DataGenerator:
        """Returns generator that iterates data stream with applied workflow.

        StopIteration from limit function will be handled here.
        """
        self._logger.debug("Iterating data stream = '%s'", id(data_stream))
        for record in data_stream:
            try:
                modified_records = self.__apply_workflow(record, workflow)
            except StopIteration as e:
                self._logger.debug("Handle StopIteration")
                modified_records = e.value

                if modified_records is not None:
                    if isinstance(modified_records, (list, tuple)):
                        yield from modified_records
                    else:  # Just one record.
                        yield modified_records

                # There is some magic.
                # It'll stop data stream and will be handled in the finally statements.
                # If you put return not under except block it will NOT work.
                return

            if modified_records is None:
                continue
            elif isinstance(modified_records, (list, tuple)):
                yield from modified_records
            else:  # Just one record.
                yield modified_records

    def __change_data(self, data_stream: DataGenerator, workflow: WorkFlow, cache: bool = False) -> DataGenerator:
        """Applies workflow for data.

        Args:
            data_stream: Data for apply workflow.
            workflow: Workflow.
            cache: Set True if you are going to write and read from the cache.

        Yields:
            obj: Generator
        """
        if cache:
            filepath = self.get_pending_cache_filepath()
            filepath.parent.mkdir(exist_ok=True)  # Create dir if does not exist.
            self._logger.debug("Recording cache file '%s'" % filepath)
            file = open(filepath, "wb")

            for modified_record in self._iterate_modified_data_stream(data_stream, workflow):
                pickle.dump(modified_record, file)
                yield modified_record

            file.close()
            rename(file.name, str(self.get_cache_filepath()))
            self._logger.debug("Cache file was created '%s'" % self.get_cache_filepath())
        else:
            yield from self._iterate_modified_data_stream(data_stream, workflow)

    def __is_cache_file_exists(self) -> bool:
        """Checks whether cache file exist."""
        path = self.get_cache_filepath()
        r = path.is_file()
        self._logger.debug("Cache file exists" if r else "Cache file doesn't exist")
        return r

    def __load_file(self, filepath: Path) -> DataGenerator:
        """Loads records from pickle file.

        Args:
            filepath: Filepath.

        Yields:
            obj: Generator records.
        """
        if not filepath.exists():
            raise FileNotFoundError(f"{filepath} doesn't exist")

        if not filepath.is_file():
            raise FileExistsError(f"{filepath} isn't file")

        if filepath.suffix != ".pickle":
            raise FileNotFoundError(f"File hasn't pickle extension")

        with open(filepath, "rb") as file:
            while True:
                try:
                    decoded_data = pickle.load(file)
                    yield decoded_data
                except EOFError:
                    break

    def _process_step(self, step: dict, record):
        res = step["callback"](record)
        self._logger.debug("    - step '%s' -> %s", step["type"], res)
        return res

    def __apply_workflow(self, record: Any, workflow: WorkFlow) -> Optional[Union[dict, List[dict]]]:
        """Creates generator records with apply workflow.

        Returns:
            obj: Generator records.

        """
        self._logger.debug("Apply workflow for %s", record)
        for step in workflow:
            if isinstance(record, (list, tuple)):
                result = []
                for r in record:
                    step_res = None
                    try:
                        step_res = self._process_step(step, r)
                    except StopIteration as e:
                        step_res = e.value
                        raise StopIteration(result if result else None)
                    finally:
                        if step_res is not None:
                            if isinstance(step_res, (list, tuple)):
                                result += step_res  # To make flat list.
                            else:
                                result.append(step_res)

                record = result
                if not record:
                    record = None
                    break  # Break iteration if step result is None.
            else:
                try:
                    record = self._process_step(step, record)
                    if record is None:
                        break  # Break workflow iteration if step result is None.
                except StopIteration:
                    raise

        self._logger.debug("-> %s", record)
        return record

    def filter(self, callback: Callable) -> "Data":
        """Append `filter` to workflow.

        Args:
            callback: Filter function.
                This function should return True or False.
                If function returns False, the record will be removed from the dataflow.

        Returns:
            Data: Data object.

        """
        self._logger.info("Apply filter")
        new_workflow = [
            {"type": "filter", "callback": lambda record: record if callback(record) else None},
        ]
        return Data(data=self, workflow=new_workflow)

    def map(self, callback: Callable) -> "Data":
        """Append `transform` function to workflow.

        Args:
            callback: Transform function.

        Returns:
            Data: Data object.

        """
        self._logger.info("Apply map")
        new_workflow = [{"type": "map", "callback": callback}]
        return Data(data=self, workflow=new_workflow)

    def _build_limit_callback(self, num) -> Callable:
        self._logger.debug("Build limit callback with limit = %s", num)

        def callback(r):
            callback.pushed += 1
            if callback.pushed == num:
                callback.pushed = 0
                self._logger.debug("Limit reached - raise StopIteration")
                raise StopIteration(r)
            else:
                return r

        callback.limit = num
        callback.pushed = 0
        return callback

    def limit(self, num: int) -> "Data":
        """Limits the stream to `num` entries.

        Args:
            num: How many records will be provided.

        Returns:
            Data: Data object.
        """
        self._logger.info("Apply limit = %s", num)
        new_workflow = [{"type": "limit", "callback": self._build_limit_callback(num)}]
        data_obj = Data(data=self, workflow=new_workflow)
        data_obj._length_hint = num
        return data_obj

    def sift(self, limit: int = None, skip: int = None) -> Generator[dict, None, None]:
        """Skips and limits records.

        Args:
            limit: Limited records.
            skip: Skipped records.

        Yields:
            Generator records.

        """
        skipped = 0
        pushed = 0

        for record in self.__load_data():
            if skip is not None and skipped < skip:
                skipped += 1
                continue
            if limit is not None and pushed == limit:
                break
            yield record
            pushed += 1

    def use_cache(self, status: bool = True) -> "Data":
        """Changes cache flag and returns self.

        Args:
            status(bool): If True the whole data stream will be saved to cache file.
            Further actions with the Data object will consume data from the cache file. True by default.

        Returns:
            Data: Data object.

        """
        self._logger.info("Cache using activated" if status else "Cache using deactivated")
        self._cache_status = status
        return self

    def find_by(self, record_field, field_values) -> Generator:
        """Get the records whose field value is written in the field_values list.

        When to use:
            You have IDs of some messages and you want get them in the stream and stop searching
            when you find all elements.

        Args:
            record_field: The record field to be searched for in the field_values list.
            field_values: List of elements among which will be searched record[record_field].

        Yields:
            dict: Generator records.

        """
        values_for_find = list(field_values)
        for record in self:
            if values_for_find:
                if record[record_field] in values_for_find:
                    values_for_find.remove(record[record_field])
                    yield record
                else:
                    continue
            else:
                break

    def write_to_file(self, file: str) -> None:
        """Writes the stream data to txt file.

        Args:
            file: Path to file.

        """
        with open(file, "w") as txt_file:
            for record in self:
                txt_file.write(f"{pprint.pformat(record)}\n" + ("-" * 50) + "\n")

    def __str__(self):
        output = "------------- Printed first 5 records -------------\n"
        for index, record in enumerate(self.__load_data()):
            if index == 5:
                break
            output += pprint.pformat(record) + "\n"
        return output

    def __bool__(self):
        for _ in self.__load_data():
            return True
        return False



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import abstractmethod

from th2_data_services.provider.interfaces.command import IHTTPProviderCommand, IGRPCProviderCommand
from th2_data_services.provider.v5.data_source.grpc import GRPCProvider5DataSource
from th2_data_services.provider.v5.data_source.http import HTTPProvider5DataSource


class IHTTPProvider5Command(IHTTPProviderCommand):
    """Interface of command for rpt-data-provider.

    Rpt-data-provider version: 5.x.y
    Protocol: HTTP
    """

    @abstractmethod
    def handle(self, data_source: HTTPProvider5DataSource):
        pass


class IGRPCProvider5Command(IGRPCProviderCommand):
    """Interface of command for rpt-data-provider.

    Rpt-data-provider version: 5.x.y
    Protocol: GRPC
    """

    @abstractmethod
    def handle(self, data_source: GRPCProvider5DataSource):
        pass



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass

from th2_data_services.provider.interfaces.struct import IEventStruct, IMessageStruct


@dataclass
class Provider5EventStruct(IEventStruct):
    """Interface for Event of data-provider v5."""

    EVENT_ID: str
    PARENT_EVENT_ID: str
    STATUS: str
    NAME: str
    TYPE: str
    BATCH_ID: str
    IS_BATCHED: str
    EVENT_TYPE: str
    END_TIMESTAMP: str
    START_TIMESTAMP: str
    ATTACHED_MESSAGES_IDS: str
    BODY: str


provider5_event_struct = Provider5EventStruct(
    EVENT_ID="eventId",
    PARENT_EVENT_ID="parentEventId",
    STATUS="successful",
    NAME="eventName",
    TYPE="type",
    BATCH_ID="batchId",
    IS_BATCHED="isBatched",
    EVENT_TYPE="eventType",
    END_TIMESTAMP="endTimestamp",
    START_TIMESTAMP="startTimestamp",
    ATTACHED_MESSAGES_IDS="attachedMessageIds",
    BODY="body",
)


@dataclass
class Provider5MessageStruct(IMessageStruct):
    """Interface for Message of data-provider v5."""

    DIRECTION: str
    SESSION_ID: str
    MESSAGE_TYPE: str
    CONNECTION_ID: str
    SESSION_ALIAS: str
    SUBSEQUENCE: str
    SEQUENCE: str
    TIMESTAMP: str
    BODY: str
    BODY_BASE64: str
    TYPE: str
    MESSAGE_ID: str
    ATTACHED_EVENT_IDS: str


provider5_message_struct = Provider5MessageStruct(
    DIRECTION="direction",
    SESSION_ID="sessionId",
    MESSAGE_TYPE="messageType",
    CONNECTION_ID="connectionId",
    SESSION_ALIAS="sessionAlias",
    SUBSEQUENCE="subsequence",
    SEQUENCE="sequence",
    TIMESTAMP="timestamp",
    BODY="body",
    BODY_BASE64="bodyBase64",
    TYPE="type",
    MESSAGE_ID="messageId",
    ATTACHED_EVENT_IDS="attachedEventIds",
)



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
from typing import Union

from th2_data_services.interfaces.adapter import IAdapter

from google.protobuf.json_format import MessageToDict
from th2_grpc_data_provider.data_provider_template_pb2 import EventData, MessageData


class GRPCObjectToDictAdapter(IAdapter):
    """GRPC Adapter decodes a GRPC object into a Dict object."""

    def handle(self, record: Union[MessageData, EventData]) -> dict:
        """Decodes MessageData or EventData as GRPC object into a Dict object.

        Args:
            record: MessageData/EventData.

        Returns:
            Dict object.
        """
        new_record = MessageToDict(record, including_default_value_fields=True)

        if isinstance(record, EventData):
            new_record["startTimestamp"] = {
                "epochSecond": record.start_timestamp.seconds,
                "nano": record.start_timestamp.nanos,
            }
            new_record["endTimestamp"] = {
                "epochSecond": record.end_timestamp.seconds,
                "nano": record.end_timestamp.nanos,
            }
        elif isinstance(record, MessageData):
            new_record["timestamp"] = {"epochSecond": record.timestamp.seconds, "nano": record.timestamp.nanos}

        try:
            new_record["body"] = json.loads(record.body)
        except (KeyError, json.JSONDecodeError):
            return new_record
        return new_record



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations
from typing import Any

from grpc._channel import _InactiveRpcError

from typing import TYPE_CHECKING

from th2_data_services.provider.exceptions import CommandError
from th2_data_services.provider.interfaces import IEventStruct, IMessageStruct

if TYPE_CHECKING:
    from th2_data_services.provider.v5.interfaces.command import IGRPCProvider5Command

from th2_data_services.provider.interfaces.data_source import IGRPCProviderDataSource

import logging

from th2_data_services.provider.interfaces.stub_builder import IEventStub, IMessageStub
from th2_data_services.provider.v5.provider_api import GRPCProvider5API
from th2_data_services.provider.v5.struct import (
    provider5_event_struct,
    provider5_message_struct,
    Provider5EventStruct,
    Provider5MessageStruct,
)
from th2_data_services.provider.v5.stub_builder import (
    provider5_event_stub_builder,
    provider5_message_stub_builder,
)

logger = logging.getLogger(__name__)


class GRPCProvider5DataSource(IGRPCProviderDataSource):
    """DataSource class which provide work with rpt-data-provider.

    Rpt-data-provider version: 5.x.y
    Protocol: GRPC
    """

    def __init__(
        self,
        url: str,
        event_struct: IEventStruct = provider5_event_struct,
        message_struct: IMessageStruct = provider5_message_struct,
        event_stub_builder: IEventStub = provider5_event_stub_builder,
        message_stub_builder: IMessageStub = provider5_message_stub_builder,
    ):
        """GRPCProvider5DataSource constructor.

        Args:
            url: Url of rpt-data-provider.
            event_struct: Event structure that is supplied by rpt-data-provider.
            message_struct: Message structure that is supplied by rpt-data-provider.
            event_stub_builder: Stub builder for broken events.
            message_stub_builder: Stub builder for broken messages.
        """
        super().__init__(
            url=url,
            event_struct=event_struct,
            message_struct=message_struct,
            event_stub_builder=event_stub_builder,
            message_stub_builder=message_stub_builder,
        )

        self.__provider_api = GRPCProvider5API(url)

        logger.info(url)

    def command(self, cmd: IGRPCProvider5Command) -> Any:
        """Execute the transmitted GRPC command.

        Args:
            cmd: GRPC Command.

        Returns:
            Any: Command response.

        Raises:
            CommandError: If the command was broken.
        """
        try:
            return cmd.handle(data_source=self)
        except _InactiveRpcError as info:
            raise CommandError(
                f"The command '{cmd.__class__.__name__}' was broken. Details of error:\n{info.details()}"
            )

    @property
    def source_api(self) -> GRPCProvider5API:
        """Returns Provider API."""
        return self.__provider_api

    @property
    def event_struct(self) -> Provider5EventStruct:
        """Returns event structure class."""
        return self._event_struct

    @property
    def message_struct(self) -> Provider5MessageStruct:
        """Returns message structure class."""
        return self._message_struct



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Interfaces for Provider Commands."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from th2_data_services.interfaces import ICommand

if TYPE_CHECKING:
    from th2_data_services.provider.interfaces.data_source import (
        IProviderDataSource,
        IHTTPProviderDataSource,
        IGRPCProviderDataSource,
    )


class IProviderCommand(ICommand):
    """Interface of command for rpt-data-provider."""

    @abstractmethod
    def handle(self, data_source: IProviderDataSource):
        pass


class IHTTPProviderCommand(IProviderCommand):
    """Interface of command for rpt-data-provider which works via HTTP."""

    @abstractmethod
    def handle(self, data_source: IHTTPProviderDataSource):
        pass


class IGRPCProviderCommand(IProviderCommand):
    """Interface of command for rpt-data-provider which works via GRPC."""

    @abstractmethod
    def handle(self, data_source: IGRPCProviderDataSource):
        pass



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import Callable

from th2_data_services.provider.interfaces.command import IProviderCommand


class ProviderAdaptableCommand(IProviderCommand):
    def __init__(self):
        """Class to make Command classes adaptable."""
        self._workflow = []

    def apply_adapter(self, adapter: Callable) -> "ProviderAdaptableCommand":
        """Adds adapter to the Command workflow.

        Note, sequence that you will add adapters make sense.

        Args:
            adapter: Callable function that will be used as adapter.

        Returns:
            self
        """
        self._workflow.append(adapter)
        return self

    def _handle_adapters(self, data):
        for step in self._workflow:
            data = step(data)
        return data



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union, Optional

from th2_data_services import Data
from th2_data_services.interfaces.events_tree import EventsTreeCollection
from th2_data_services.events_tree.exceptions import FieldIsNotExist
from th2_data_services.provider.interfaces.struct import IEventStruct
from th2_data_services.provider.v5.data_source import HTTPProvider5DataSource, GRPCProvider5DataSource
from th2_data_services.provider.v5.struct import provider5_event_struct


class EventsTreeCollectionProvider5(EventsTreeCollection):
    """EventsTreesCollections for data-provider v5."""

    def __init__(
        self,
        data: Data,
        data_source: Union[GRPCProvider5DataSource, HTTPProvider5DataSource] = None,
        preserve_body: bool = False,
        event_struct: IEventStruct = provider5_event_struct,
        stub: bool = False,
    ):
        """EventsTreeCollectionProvider5 constructor.

        Args:
            data: Data object.
            data_source: Data Source object.
            preserve_body: If True it will preserve 'body' field in the Events.
            event_struct: Event struct object.
            stub: If True it will create stub when event is broken.
        """
        self._event_struct = event_struct  # Should be placed before super!

        super().__init__(data=data, data_source=data_source, preserve_body=preserve_body, stub=stub)

    def _get_event_id(self, event) -> str:
        """Gets event id from the event.

        Returns:
            Event id.

        Raises:
            FieldIsNotExist: If the event doesn't have an 'event id' field.
        """
        try:
            return event[self._event_struct.EVENT_ID]
        except KeyError:
            raise FieldIsNotExist(self._event_struct.EVENT_ID)

    def _get_event_name(self, event) -> str:
        """Gets event name from the event.

        Returns:
            Event name.

        Raises:
            FieldIsNotExist: If the event doesn't have an 'event name' field.
        """
        try:
            return event[self._event_struct.NAME]
        except KeyError:
            raise FieldIsNotExist(self._event_struct.NAME)

    def _get_parent_event_id(self, event) -> Optional[str]:
        """Gets parent event id from event.

        Returns:
            Parent event id.
        """
        return event.get(self._event_struct.PARENT_EVENT_ID)



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


class EventIdNotInTree(Exception):
    def __init__(self, id_):
        """Exception for the case when the tree hasn't the event.

        Args:
            id_: Event id.
        """
        self._id = id_

    def __str__(self):
        return f"Event with the id '{self._id}' doesn't exist in the tree"


class FieldIsNotExist(Exception):
    def __init__(self, field_name):
        """Exception for the case when event as dict hasn't field.

        Args:
            field_name: Field name.
        """
        self._field_name = field_name

    def __str__(self):
        return f"Event doesn't have '{self._field_name}' field"



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from sseclient import Event as SSEEvent
from urllib3.exceptions import HTTPError
import json

from th2_data_services.interfaces import IAdapter


class SSEAdapter(IAdapter):
    """SSE Adapter handles bytes from sse-stream into Dict object."""

    def handle(self, record: SSEEvent) -> dict:
        """Adapter handler.

        Args:
            record: SSE Event.

        Returns:
            Dict object.
        """
        if record.event == "error":
            raise HTTPError(record.data)
        if record.event not in ["close", "keep_alive", "message_ids"]:
            try:
                return json.loads(record.data)
            except json.JSONDecodeError as e:
                raise ValueError(f"json.decoder.JSONDecodeError: Invalid json received.\n" f"{e}\n" f"{record.data}")



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union, Type

from th2_data_services.provider.interfaces import IProviderDataSource
from th2_data_services.provider.v5.commands.grpc import GetEventsById as GetEventsByIdFromGRPC
from th2_data_services.provider.v5.commands.http import GetEventsById as GetEventsByIdFromHTTP

from th2_data_services.provider.v5.commands.grpc import GetEventById as GetEventByIdFromGRPC
from th2_data_services.provider.v5.commands.http import GetEventById as GetEventByIdFromHTTP

from th2_data_services.provider.v5.data_source.grpc import GRPCProvider5DataSource
from th2_data_services.provider.v5.data_source.http import HTTPProvider5DataSource


def resolver_get_event_by_id(
    data_source: IProviderDataSource,
) -> Union[Type[GetEventByIdFromHTTP], Type[GetEventByIdFromGRPC]]:
    """Resolves what 'GetEventById' command you need to use based Data Source.

    Args:
        data_source: DataSource instance.

    Returns:
        GetEventById command.
    """
    if isinstance(data_source, GRPCProvider5DataSource):
        return GetEventByIdFromGRPC
    elif isinstance(data_source, HTTPProvider5DataSource):
        return GetEventByIdFromHTTP
    else:
        raise ValueError("Unknown DataSource Object")


def resolver_get_events_by_id(
    data_source: IProviderDataSource,
) -> Union[Type[GetEventsByIdFromHTTP], Type[GetEventsByIdFromGRPC]]:
    """Resolves what 'GetEventsById' command you need to use based Data Source.

    Args:
        data_source: DataSource instance.

    Returns:
        GetEventsById command.
    """
    if isinstance(data_source, GRPCProvider5DataSource):
        return GetEventsByIdFromGRPC
    elif isinstance(data_source, HTTPProvider5DataSource):
        return GetEventsByIdFromHTTP
    else:
        raise ValueError("Unknown DataSource Object")



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Generator, List
import json
import simplejson
from datetime import datetime, timezone
from functools import partial

from th2_data_services import Filter, Data
from th2_data_services.provider.exceptions import EventNotFound, MessageNotFound
from th2_data_services.provider.v5.interfaces.command import IHTTPProvider5Command
from th2_data_services.provider.v5.data_source.http import HTTPProvider5DataSource
from th2_data_services.provider.v5.provider_api import HTTPProvider5API
from th2_data_services.provider.command import ProviderAdaptableCommand
from th2_data_services.sse_client import SSEClient
from th2_data_services.provider.adapters.adapter_sse import SSEAdapter
from th2_data_services.decode_error_handler import UNICODE_REPLACE_HANDLER

import logging

logger = logging.getLogger(__name__)


class GetEventById(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It retrieves the event by id.

    Returns:
        dict: Th2 event.

    Raises:
        EventNotFound: If event by Id wasn't found.
    """

    def __init__(self, id: str, use_stub=False):
        """GetEventById constructor.

        Args:
            id: Event id.
            use_stub: If True the command returns stub instead of exception.

        """
        super().__init__()
        self._id = id
        self._stub_status = use_stub

    def handle(self, data_source: HTTPProvider5DataSource) -> dict:  # noqa: D102
        api: HTTPProvider5API = data_source.source_api
        url = api.get_url_find_event_by_id(self._id)

        logger.info(url)

        response = api.execute_request(url)
        try:
            event = response.json()
        except (json.JSONDecodeError, simplejson.JSONDecodeError):
            if self._stub_status:
                return data_source.event_stub_builder.build({data_source.event_struct.EVENT_ID: self._id})
            else:
                logger.error(f"Unable to find the message. Id: {self._id}")
                raise EventNotFound(self._id)
        return self._handle_adapters(event)


class GetEventsById(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It retrieves the events by ids.

    Returns:
        List[dict]: Th2 events.

    Raises:
        EventNotFound: If any event by Id wasn't found.
    """

    def __init__(self, ids: List[str], use_stub=False):
        """GetEventsById constructor.

        Args:
            ids: Event id list.
            use_stub: If True the command returns stub instead of exception.

        """
        super().__init__()
        self._ids: ids = ids
        self._stub_status = use_stub

    def handle(self, data_source: HTTPProvider5DataSource):  # noqa: D102
        result = []
        for event_id in self._ids:
            event = GetEventById(event_id, use_stub=self._stub_status).handle(data_source)
            result.append(self._handle_adapters(event))

        return result


class GetEventsSSEBytes(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches events stream by options.

    Returns:
        Iterable[dict]: Stream of Th2 events.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        end_timestamp: datetime = None,
        parent_event: str = None,
        search_direction: str = "next",
        resume_from_id: str = None,
        result_count_limit: int = None,
        keep_open: bool = False,
        limit_for_parent: int = None,
        attached_messages: bool = False,
        filters: (Filter, List[Filter]) = None,
    ):
        """GetEventsSSEBytes constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            parent_event: Match events to the specified parent.
            search_direction: Search direction.
            resume_from_id: Event id from which search starts.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
                the one closest to the specified timestamp.
            limit_for_parent: How many children events for each parent do we want to request.
            attached_messages: Gets messages ids which linked to events.
            filters: Filters using in search for messages.

        """
        super().__init__()
        self._start_timestamp = int(1000 * start_timestamp.replace(tzinfo=timezone.utc).timestamp())
        self._end_timestamp = int(1000 * end_timestamp.replace(tzinfo=timezone.utc).timestamp())
        self._parent_event = parent_event
        self._search_direction = search_direction
        self._resume_from_id = resume_from_id
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._limit_for_parent = limit_for_parent
        self._metadata_only = False
        self._attached_messages = attached_messages
        self._filters = filters
        if isinstance(filters, Filter):
            self._filters = filters.url()
        elif isinstance(filters, (tuple, list)):
            self._filters = "".join([filter_.url() for filter_ in filters])

    def handle(self, data_source: HTTPProvider5DataSource):  # noqa: D102
        api: HTTPProvider5API = data_source.source_api
        url = api.get_url_search_sse_events(
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            parent_event=self._parent_event,
            search_direction=self._search_direction,
            resume_from_id=self._resume_from_id,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            limit_for_parent=self._limit_for_parent,
            metadata_only=self._metadata_only,
            attached_messages=self._attached_messages,
            filters=self._filters,
        )

        logger.info(url)

        for response in api.execute_sse_request(url):
            response = self._handle_adapters(response)
            if response is not None:
                yield response


class GetEventsSSEEvents(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches events stream by options.

    Returns:
        Iterable[dict]: Stream of Th2 events.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        end_timestamp: datetime = None,
        parent_event: str = None,
        search_direction: str = "next",
        resume_from_id: str = None,
        result_count_limit: int = None,
        keep_open: bool = False,
        limit_for_parent: int = None,
        attached_messages: bool = False,
        filters: (Filter, List[Filter]) = None,
        char_enc: str = "utf-8",
        decode_error_handler: str = UNICODE_REPLACE_HANDLER,
    ):
        """GetEventsSSEEvents constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            parent_event: Match events to the specified parent.
            search_direction: Search direction.
            resume_from_id: Event id from which search starts.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
                the one closest to the specified timestamp.
            limit_for_parent: How many children events for each parent do we want to request.
            attached_messages: Gets messages ids which linked to events.
            filters: Filters using in search for messages.
            char_enc: Encoding for the byte stream.
            decode_error_handler: Registered decode error handler.

        """
        super().__init__()
        self._start_timestamp = start_timestamp
        self._end_timestamp = end_timestamp
        self._parent_event = parent_event
        self._search_direction = search_direction
        self._resume_from_id = resume_from_id
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._limit_for_parent = limit_for_parent
        self._attached_messages = attached_messages
        self._filters = filters
        self._char_enc = char_enc
        self._decode_error_handler = decode_error_handler

    def handle(self, data_source: HTTPProvider5DataSource):  # noqa: D102
        response = GetEventsSSEBytes(
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            parent_event=self._parent_event,
            search_direction=self._search_direction,
            resume_from_id=self._resume_from_id,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            limit_for_parent=self._limit_for_parent,
            attached_messages=self._attached_messages,
            filters=self._filters,
        ).handle(data_source)
        client = SSEClient(response, char_enc=self._char_enc, decode_errors_handler=self._decode_error_handler)
        for record in client.events():
            record = self._handle_adapters(record)
            if record is not None:
                yield record


class GetEvents(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches events stream by options.

    Returns:
        Iterable[dict]: Stream of Th2 events.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        end_timestamp: datetime = None,
        parent_event: str = None,
        search_direction: str = "next",
        resume_from_id: str = None,
        result_count_limit: int = None,
        keep_open: bool = False,
        limit_for_parent: int = None,
        attached_messages: bool = False,
        filters: (Filter, List[Filter]) = None,
        cache: bool = False,
    ):
        """GetEvents constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            parent_event: Match events to the specified parent.
            search_direction: Search direction.
            resume_from_id: Event id from which search starts.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
                the one closest to the specified timestamp.
            limit_for_parent: How many children events for each parent do we want to request.
            attached_messages: Gets messages ids which linked to events.
            filters: Filters using in search for messages.
            cache: If True, all requested data from rpt-data-provider will be saved to cache.

        """
        super().__init__()
        self._start_timestamp = start_timestamp
        self._end_timestamp = end_timestamp
        self._parent_event = parent_event
        self._search_direction = search_direction
        self._resume_from_id = resume_from_id
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._limit_for_parent = limit_for_parent
        self._attached_messages = attached_messages
        self._filters = filters
        self._cache = cache

    def handle(self, data_source: HTTPProvider5DataSource) -> Data:  # noqa: D102
        source = partial(self.__handle_stream, data_source)
        adapter = SSEAdapter()
        return Data(source).map(adapter.handle).use_cache(self._cache)

    def __handle_stream(self, data_source: HTTPProvider5DataSource) -> Generator[dict, None, None]:
        stream = GetEventsSSEEvents(
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            parent_event=self._parent_event,
            search_direction=self._search_direction,
            resume_from_id=self._resume_from_id,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            limit_for_parent=self._limit_for_parent,
            attached_messages=self._attached_messages,
            filters=self._filters,
        ).handle(data_source)

        for event in stream:
            event = self._handle_adapters(event)
            yield event


class GetMessageById(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It retrieves the message by id.

    Returns:
        dict: Th2 message.

    Raises:
        MessageNotFound: If message by id wasn't found.
    """

    def __init__(self, id: str, use_stub=False):
        """GetMessageById constructor.

        Args:
            id: Message id.
            use_stub: If True the command returns stub instead of exception.

        """
        super().__init__()
        self._id = id
        self._stub_status = use_stub

    def handle(self, data_source: HTTPProvider5DataSource) -> dict:  # noqa: D102
        api: HTTPProvider5API = data_source.source_api
        url = api.get_url_find_message_by_id(self._id)

        logger.info(url)

        response = api.execute_request(url)
        try:
            message = response.json()
        except (json.JSONDecodeError, simplejson.JSONDecodeError):
            if self._stub_status:
                return data_source.message_stub_builder.build({data_source.message_struct.MESSAGE_ID: self._id})
            else:
                exception_msg = f"Unable to find the message. Id: {self._id}"
                logger.error(exception_msg)
                raise MessageNotFound(self._id)
        return self._handle_adapters(message)


class GetMessagesById(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It retrieves the messages by ids.

    Returns:
        List[dict]: Th2 messages.

    Raises:
        MessageNotFound: If any message by id wasn't found.
    """

    def __init__(self, ids: List[str], use_stub=False):
        """GetMessagesById constructor.

        Args:
            ids: Message id list.
            use_stub: If True the command returns stub instead of exception.

        """
        super().__init__()
        self._ids: ids = ids
        self._stub_status = use_stub

    def handle(self, data_source: HTTPProvider5DataSource) -> List[dict]:  # noqa: D102
        result = []
        for message_id in self._ids:
            message = GetMessageById(message_id, use_stub=self._stub_status).handle(data_source)
            result.append(self._handle_adapters(message))

        return result


class GetMessagesSSEBytes(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches messages stream by options.

    Returns:
        Iterable[dict]: Stream of Th2 messages.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        stream: List[str],
        end_timestamp: datetime = None,
        resume_from_id: str = None,
        search_direction: str = "next",
        result_count_limit: int = None,
        keep_open: bool = False,
        message_id: List[str] = None,
        attached_events: bool = False,
        lookup_limit_days: int = None,
        filters: (Filter, List[Filter]) = None,
    ):
        """GetMessagesSSEBytes constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            stream: Alias of messages.
            resume_from_id: Message id from which search starts.
            search_direction: Search direction.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
            message_id: List of message IDs to restore search. If given, it has
                the highest priority and ignores stream (uses streams from ids), startTimestamp and resumeFromId.
            attached_events: If true, additionally load attached_event_ids
            lookup_limit_days: The number of days that will be viewed on
                the first request to get the one closest to the specified timestamp.
            filters: Filters using in search for messages.
        """
        super().__init__()
        self._start_timestamp = int(1000 * start_timestamp.replace(tzinfo=timezone.utc).timestamp())
        self._end_timestamp = (
            end_timestamp
            if end_timestamp is None
            else int(1000 * end_timestamp.replace(tzinfo=timezone.utc).timestamp())
        )
        self._stream = stream
        self._resume_from_id = resume_from_id
        self._search_direction = search_direction
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._message_id = message_id
        self._attached_events = attached_events
        self._lookup_limit_days = lookup_limit_days
        self._filters = filters
        if isinstance(filters, Filter):
            self._filters = filters.url()
        elif isinstance(filters, (tuple, list)):
            self._filters = "".join([filter_.url() for filter_ in filters])

    def handle(self, data_source: HTTPProvider5DataSource) -> Generator[dict, None, None]:  # noqa: D102
        api: HTTPProvider5API = data_source.source_api
        url = api.get_url_search_sse_messages(
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            stream=[""],
            resume_from_id=self._resume_from_id,
            search_direction=self._search_direction,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            attached_events=self._attached_events,
            lookup_limit_days=self._lookup_limit_days,
            filters=self._filters,
        ).replace("&stream=", "")

        logger.info(url)

        fixed_part_len = len(url)
        current_url, resulting_urls = "", []
        for stream in self._stream:
            stream = f"&stream={stream}"
            if fixed_part_len + len(current_url) + len(stream) >= 2048:
                resulting_urls.append(url + current_url)
                current_url = ""
            current_url += stream
        if current_url:
            resulting_urls.append(url + current_url)

        for url in resulting_urls:
            for response in api.execute_sse_request(url):
                response = self._handle_adapters(response)
                if response is not None:
                    yield response


class GetMessagesSSEEvents(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches messages stream by options.

    Returns:
        Iterable[dict]: Stream of Th2 messages.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        stream: List[str],
        end_timestamp: datetime = None,
        resume_from_id: str = None,
        search_direction: str = "next",
        result_count_limit: int = None,
        keep_open: bool = False,
        message_id: List[str] = None,
        attached_events: bool = False,
        lookup_limit_days: int = None,
        filters: (Filter, List[Filter]) = None,
        char_enc: str = "utf-8",
        decode_error_handler: str = UNICODE_REPLACE_HANDLER,
    ):
        """GetMessagesSSEEvents constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            stream: Alias of messages.
            resume_from_id: Message id from which search starts.
            search_direction: Search direction.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
            message_id: List of message IDs to restore search. If given, it has
                the highest priority and ignores stream (uses streams from ids), startTimestamp and resumeFromId.
            attached_events: If true, additionally load attached_event_ids
            lookup_limit_days: The number of days that will be viewed on
                the first request to get the one closest to the specified timestamp.
            filters: Filters using in search for messages.
            char_enc: Character encode that will use SSEClient.
            decode_error_handler: Decode error handler.
        """
        super().__init__()
        self._start_timestamp = start_timestamp
        self._end_timestamp = end_timestamp
        self._stream = stream
        self._resume_from_id = resume_from_id
        self._search_direction = search_direction
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._message_id = message_id
        self._attached_events = attached_events
        self._lookup_limit_days = lookup_limit_days
        self._filters = filters
        self._char_enc = char_enc
        self._decode_error_handler = decode_error_handler

    def handle(self, data_source: HTTPProvider5DataSource) -> Generator[dict, None, None]:  # noqa: D102
        response = GetMessagesSSEBytes(
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            stream=self._stream,
            resume_from_id=self._resume_from_id,
            search_direction=self._search_direction,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            attached_events=self._attached_events,
            lookup_limit_days=self._lookup_limit_days,
            filters=self._filters,
        ).handle(data_source)

        client = SSEClient(response, char_enc=self._char_enc, decode_errors_handler=self._decode_error_handler)

        for record in client.events():
            record = self._handle_adapters(record)
            if record is not None:
                yield record


class GetMessages(IHTTPProvider5Command, ProviderAdaptableCommand):
    """A Class-Command for request to rpt-data-provider.

    It searches messages stream by options.

    Returns:
        Iterable[dict]: Stream of Th2 messages.
    """

    def __init__(
        self,
        start_timestamp: datetime,
        stream: List[str],
        end_timestamp: datetime = None,
        resume_from_id: str = None,
        search_direction: str = "next",
        result_count_limit: int = None,
        keep_open: bool = False,
        message_id: List[str] = None,
        attached_events: bool = False,
        lookup_limit_days: int = None,
        filters: (Filter, List[Filter]) = None,
        char_enc: str = "utf-8",
        decode_error_handler: str = UNICODE_REPLACE_HANDLER,
        cache: bool = False,
    ):
        """GetMessages constructor.

        Args:
            start_timestamp: Start timestamp of search.
            end_timestamp: End timestamp of search.
            stream: Alias of messages.
            resume_from_id: Message id from which search starts.
            search_direction: Search direction.
            result_count_limit: Result count limit.
            keep_open: If the search has reached the current moment.
                It is need to wait further for the appearance of new data.
            message_id: List of message IDs to restore search. If given, it has
                the highest priority and ignores stream (uses streams from ids), startTimestamp and resumeFromId.
            attached_events: If true, additionally load attached_event_ids
            lookup_limit_days: The number of days that will be viewed on
                the first request to get the one closest to the specified timestamp.
            filters: Filters using in search for messages.
            char_enc: Encoding for the byte stream.
            decode_error_handler: Registered decode error handler.
            cache: If True, all requested data from rpt-data-provider will be saved to cache.
        """
        super().__init__()
        self._start_timestamp = start_timestamp
        self._end_timestamp = end_timestamp
        self._stream = stream
        self._resume_from_id = resume_from_id
        self._search_direction = search_direction
        self._result_count_limit = result_count_limit
        self._keep_open = keep_open
        self._message_id = message_id
        self._attached_events = attached_events
        self._lookup_limit_days = lookup_limit_days
        self._filters = filters
        self._char_enc = char_enc
        self._decode_error_handler = decode_error_handler
        self._cache = cache

    def handle(self, data_source: HTTPProvider5DataSource) -> Data:  # noqa: D102
        source = partial(self.__handle_stream, data_source)
        adapter = SSEAdapter()
        return Data(source).map(adapter.handle).use_cache(self._cache)

    def __handle_stream(self, data_source: HTTPProvider5DataSource) -> Generator[dict, None, None]:
        stream = GetMessagesSSEEvents(
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            stream=self._stream,
            resume_from_id=self._resume_from_id,
            search_direction=self._search_direction,
            result_count_limit=self._result_count_limit,
            keep_open=self._keep_open,
            attached_events=self._attached_events,
            lookup_limit_days=self._lookup_limit_days,
            filters=self._filters,
        ).handle(data_source)

        for message in stream:
            message = self._handle_adapters(message)
            yield message



from typing import Union, List, Tuple

from th2_grpc_data_provider.data_provider_template_pb2 import Filter as grpc_Filter, FilterName as grpc_FilterName
import google.protobuf.wrappers_pb2


class Filter:
    """The class for using rpt-data-provider filters API."""

    def __init__(
        self, name: str, values: Union[List[str], Tuple[str], str], negative: bool = False, conjunct: bool = False
    ):
        """Filter constructor.

        Args:
            name (str): Filter name.
            values (Union[List[str], Tuple[str], str]): One string with filter value or list of filter values.
            negative (bool):  If true, will match events/messages that do not match those specified values.
                If false, will match the events/messages by their values. Defaults to false.
            conjunct (bool): If true, each of the specific filter values should be applied
                If false, at least one of the specific filter values must be applied.
        """
        self.name = name

        if isinstance(values, (list, tuple)):
            self.values = map(str, values)
        else:
            self.values = [str(values)]

        self.negative = negative
        self.conjunct = conjunct

    def url(self) -> str:
        """Forms a filter.

        For help use this readme:
        https://github.com/th2-net/th2-rpt-data-provider#filters-api.

        Returns:
            str: Formed filter.
        """
        return (
            f"&filters={self.name}"
            + "".join([f"&{self.name}-values={val}" for val in self.values])
            + f"&{self.name}-negative={self.negative}"
        )

    def grpc(self) -> grpc_Filter:
        return grpc_Filter(
            name=grpc_FilterName(filter_name=self.name),
            negative=google.protobuf.wrappers_pb2.BoolValue(value=self.negative),
            values=self.values,
            conjunct=google.protobuf.wrappers_pb2.BoolValue(value=self.conjunct),
        )



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from th2_data_services.provider.interfaces.stub_builder import IEventStub, IMessageStub
from th2_data_services.provider.v5.struct import (
    provider5_event_struct,
    provider5_message_struct,
)


class Provider5EventStubBuilder(IEventStub):
    def __init__(self, event_struct=provider5_event_struct):
        """Event stub builder for Provider v5.

        Args:
            event_struct: Event struct class.
        """
        self.event_fields = event_struct
        super().__init__()  # Requirement to define fields for the template earlier.

    @property
    def template(self) -> dict:
        """Event stub template.

        Returns:
            (dict) Event stub template.
        """
        return {
            self.event_fields.ATTACHED_MESSAGES_IDS: [],
            self.event_fields.BATCH_ID: "Broken_Event",
            self.event_fields.END_TIMESTAMP: {"nano": 0, "epochSecond": 0},
            self.event_fields.START_TIMESTAMP: {"nano": 0, "epochSecond": 0},
            self.event_fields.TYPE: "event",
            self.event_fields.EVENT_ID: self.REQUIRED_FIELD,
            self.event_fields.NAME: "Broken_Event",
            self.event_fields.EVENT_TYPE: "Broken_Event",
            self.event_fields.PARENT_EVENT_ID: "Broken_Event",
            self.event_fields.STATUS: None,
            self.event_fields.IS_BATCHED: None,
        }


class Provider5MessageStubBuilder(IMessageStub):
    def __init__(self, message_struct=provider5_message_struct):
        """Event stub builder for Provider v5.

        Args:
            message_struct: Message struct class.
        """
        self.message_fields = message_struct
        super().__init__()  # Requirement to define fields for the template earlier.

    @property
    def template(self) -> dict:
        """Message stub template.

        Returns:
            (dict) Message stub template.
        """
        return {
            self.message_fields.DIRECTION: None,
            self.message_fields.SESSION_ID: "Broken_Message",
            self.message_fields.MESSAGE_TYPE: "Broken_Message",
            self.message_fields.TIMESTAMP: {"nano": 0, "epochSecond": 0},
            self.message_fields.BODY: [],
            self.message_fields.BODY_BASE64: [],
            self.message_fields.TYPE: "message",
            self.message_fields.MESSAGE_ID: self.REQUIRED_FIELD,
            self.message_fields.ATTACHED_EVENT_IDS: [],
        }


provider5_event_stub_builder = Provider5EventStubBuilder()
provider5_message_stub_builder = Provider5MessageStubBuilder()



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json

from setuptools import setup, find_packages

with open("package_info.json", "r") as file:
    package_info = json.load(file)

package_name = package_info["package_name"].replace("-", "_")
package_version = package_info["package_version"]

with open("README.md", "r") as file:
    long_description = file.read()

with open("requirements.txt", "r") as file:
    requirements = [line.strip() for line in file.readlines() if not line.startswith("#") and line != "\n"]

setup(
    name=package_name,
    version=package_version,
    description=package_name,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="TH2-devs",
    author_email="th2-devs@exactprosystems.com",
    url="https://github.com/th2-net/th2-data-services",
    license="Apache License 2.0",
    python_requires=">=3.7",
    install_requires=requirements,
    packages=find_packages(include=["th2_data_services", "th2_data_services.*"]),
    include_package_data=True,
)



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from collections import namedtuple
from typing import Iterable, List, Optional

from google.protobuf.empty_pb2 import Empty
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.wrappers_pb2 import BoolValue, Int32Value, Int64Value
from th2_grpc_common.common_pb2 import MessageID, EventID, ConnectionID, Direction
from th2_grpc_data_provider.data_provider_template_pb2_grpc import DataProviderStub
from th2_grpc_data_provider.data_provider_template_pb2 import (
    EventSearchRequest,
    MessageSearchRequest,
    FilterName,
    MatchRequest,
    FilterInfo,
    ListFilterName,
    StreamResponse,
    EventData,
    StringList,
    IsMatched,
    MessageData,
    TimeRelation,
    Filter,
)
from grpc import Channel, insecure_channel
from th2_data_services.provider.interfaces.source_api import IGRPCProviderSourceAPI

logger = logging.getLogger(__name__)

BasicRequest = namedtuple(
    "BasicRequest",
    ["start_timestamp", "end_timestamp", "result_count_limit", "keep_open", "search_direction", "filters"],
)


class GRPCProvider5API(IGRPCProviderSourceAPI):
    def __init__(self, url: str):
        """GRPC Provider5 API.

        Args:
            url: GRPC data source url.
        """
        self._create_connection(url)

    def _create_connection(self, url: str) -> None:
        """Creates gRPC channel to gRPC-server.

        Args:
            url: Url of gRPC-server.
        """
        channel: Channel = insecure_channel(url)
        self.__stub: DataProviderStub = DataProviderStub(channel)

    def get_message_streams(self) -> StringList:
        """GRPC-API `getMessageStreams` call returns a list of message stream names."""
        return self.__stub.getMessageStreams(Empty())

    def search_events(
        self,
        start_timestamp: int = None,
        end_timestamp: int = None,
        parent_event: str = None,
        search_direction: str = "NEXT",
        resume_from_id: str = None,
        result_count_limit: int = None,
        keep_open: bool = False,
        limit_for_parent: int = None,
        metadata_only: bool = True,
        attached_messages: bool = False,
        filters: Optional[List[Filter]] = None,
    ) -> Iterable[StreamResponse]:
        """GRPC-API `searchEvents` call creates an event or an event metadata stream that matches the filter.

        Args:
            start_timestamp: Sets the search starting point. Expected in nanoseconds. One of the 'start_timestamp'
                or 'resume_from_id' must not absent.
            end_timestamp: Sets the timestamp to which the search will be performed, starting with 'start_timestamp'.
                Expected in nanoseconds.
            parent_event: Match events to the specified parent.
            search_direction: Sets the lookup direction. Can be 'NEXT' or 'PREVIOUS'.
            resume_from_id: The last event id from which we start searching for events.
            result_count_limit: Sets the maximum amount of events to return.
            keep_open: Option if the search has reached the current moment,
                it is necessary to wait further for the appearance of new data.
            limit_for_parent: How many children events for each parent do we want to request.
            metadata_only: Receive only metadata (true) or entire event (false) (without attachedMessageIds).
            attached_messages: Option if you want to load attachedMessageIds additionally.
            filters: Which filters to apply in a search.

        Returns:
            Iterable object which return events as parts of streaming response.
        """
        self.__search_basic_checks(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            result_count_limit=result_count_limit,
            resume_from_ids=resume_from_id,
            search_direction=search_direction,
        )

        basic_request = self.__build_basic_request_object(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            search_direction=search_direction,
            result_count_limit=result_count_limit,
            keep_open=keep_open,
            filters=filters,
        )
        parent_event = EventID(id=parent_event) if parent_event else None
        resume_from_id = EventID(id=resume_from_id) if resume_from_id else None
        limit_for_parent = Int64Value(value=limit_for_parent) if limit_for_parent else None
        metadata_only = BoolValue(value=metadata_only)
        attached_messages = BoolValue(value=attached_messages)

        event_search_request = EventSearchRequest(
            start_timestamp=basic_request.start_timestamp,
            end_timestamp=basic_request.end_timestamp,
            parent_event=parent_event,
            search_direction=basic_request.search_direction,
            resume_from_id=resume_from_id,
            result_count_limit=basic_request.result_count_limit,
            keep_open=basic_request.keep_open,
            limit_for_parent=limit_for_parent,
            metadata_only=metadata_only,
            attached_messages=attached_messages,
            filters=basic_request.filters,
        )
        return self.__stub.searchEvents(event_search_request)

    def search_messages(
        self,
        start_timestamp: int,
        stream: List[str],
        end_timestamp: int = None,
        resume_from_id: str = None,
        search_direction: str = "NEXT",
        result_count_limit: int = None,
        keep_open: bool = False,
        filters: Optional[List[Filter]] = None,
    ) -> Iterable[StreamResponse]:
        """GRPC-API `searchMessages` call creates a message stream that matches the filter.

        Args:
            start_timestamp: Sets the search starting point. Expected in nanoseconds. One of the 'start_timestamp'
                or 'resume_from_id' must not absent.
            stream: Sets the stream ids to search in.
            end_timestamp: Sets the timestamp to which the search will be performed, starting with 'start_timestamp'.
                Expected in nanoseconds.
            search_direction: Sets the lookup direction. Can be 'NEXT' or 'PREVIOUS'.
            resume_from_id: The last event id from which we start searching for messages.
            result_count_limit: Sets the maximum amount of messages to return.
            keep_open: Option if the search has reached the current moment,
                it is necessary to wait further for the appearance of new data.
            filters: Which filters to apply in a search.

        Returns:
            Iterable object which return messages as parts of streaming response.
        """
        if stream is None:
            raise TypeError("Argument 'stream' is required.")
        self.__search_basic_checks(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            result_count_limit=result_count_limit,
            resume_from_ids=resume_from_id,
            search_direction=search_direction,
        )

        basic_request = self.__build_basic_request_object(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            result_count_limit=result_count_limit,
            keep_open=keep_open,
            search_direction=search_direction,
            filters=filters,
        )
        resume_from_id = self.__build_message_id_object(resume_from_id) if resume_from_id else None

        message_search_request = MessageSearchRequest(
            start_timestamp=basic_request.start_timestamp,
            end_timestamp=basic_request.end_timestamp,
            stream=StringList(list_string=stream),
            search_direction=basic_request.search_direction,
            resume_from_id=resume_from_id,
            result_count_limit=basic_request.result_count_limit,
            keep_open=basic_request.keep_open,
            filters=basic_request.filters,
        )
        return self.__stub.searchMessages(message_search_request)

    @staticmethod
    def __search_basic_checks(
        start_timestamp: Optional[int],
        end_timestamp: Optional[int],
        resume_from_ids: Optional[str],
        search_direction: Optional[str],
        result_count_limit: Optional[int],
    ):
        if start_timestamp is None and resume_from_ids is None:
            raise ValueError("One of the 'startTimestamp' or 'resumeFromId(s)' must not be None.")

        if end_timestamp is None and result_count_limit is None:
            raise ValueError("One of the 'end_timestamp' or 'result_count_limit' must not be None.")

        if (
            start_timestamp is not None
            and len(str(start_timestamp)) != 19
            or end_timestamp is not None
            and len(str(end_timestamp)) != 19
        ):
            raise ValueError("Arguments 'start_timestamp' and 'end_timestamp' are expected in nanoseconds.")

        if search_direction is not None:
            search_direction = search_direction.upper()
            if search_direction not in ("NEXT", "PREVIOUS"):
                raise ValueError("Argument 'search_direction' must be 'NEXT' or 'PREVIOUS'.")
        else:
            raise ValueError("Argument 'search_direction' must be 'NEXT' or 'PREVIOUS'.")

    def __build_basic_request_object(
        self,
        start_timestamp: int = None,
        end_timestamp: int = None,
        result_count_limit: int = None,
        keep_open: bool = False,
        search_direction: str = "NEXT",
        filters: Optional[List[Filter]] = None,
    ) -> BasicRequest:
        """Builds a BasicRequest wrapper-object.

        Args:
            start_timestamp: Start Timestamp for request.
            end_timestamp: End Timestamp for request.
            result_count_limit: Data count limit.
            keep_open: Option for stream-request.
            search_direction: Searching direction.
            filters: Which filters to apply in a request.

        Returns:
            BasicRequest wrapper-object.
        """
        if filters is None:
            filters = []

        start_timestamp = self.__build_timestamp_object(start_timestamp) if start_timestamp else None
        end_timestamp = self.__build_timestamp_object(end_timestamp) if end_timestamp else None
        search_direction = TimeRelation.Value(search_direction)  # getting a value from enum
        result_count_limit = Int32Value(value=result_count_limit) if result_count_limit else None
        keep_open = BoolValue(value=keep_open)

        basic_request = BasicRequest(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            search_direction=search_direction,
            result_count_limit=result_count_limit,
            keep_open=keep_open,
            filters=filters,
        )
        return basic_request

    @staticmethod
    def __build_timestamp_object(timestamp: int) -> Timestamp:
        """Builds a Timestamp of 'protobuf' entity.

        Args:
            timestamp: Timestamp in nanoseconds.

        Returns:
            Timestamp object.
        """
        nanos = timestamp % 10 ** 9
        seconds = timestamp // 10 ** 9
        timestamp = Timestamp(seconds=seconds, nanos=nanos)
        return timestamp

    @staticmethod
    def __build_message_id_object(message_id: str) -> MessageID:
        """Builds a MessageID of 'protobuf' entity.

        Args:
            message_id: Message id.

        Returns:
            MessageID object.
        """
        split_id = message_id.split(":")
        split_id.reverse()  # Parse the message id from the end.

        sequence, direction, alias = split_id[0], split_id[1].upper(), split_id[2:]
        split_sequence = sequence.split(".")
        sequence, subsequence = int(split_sequence[0]), split_sequence[1:]
        subsequence = map(int, subsequence) if subsequence else subsequence
        alias.reverse()
        alias = ":".join(alias)

        connection_id = ConnectionID(session_alias=alias)
        direction = Direction.Value(direction)  # Get a value from enum.

        message_id = MessageID(
            connection_id=connection_id,
            direction=direction,
            sequence=sequence,
            subsequence=subsequence,
        )
        return message_id

    def get_event(self, event_id: str) -> EventData:
        """GRPC-API `getEvent` call returns a single event with the specified id."""
        event_id = EventID(id=event_id)
        return self.__stub.getEvent(event_id)

    def get_message(self, message_id: str) -> MessageData:
        """GRPC-API `getMessage` call returns a single message with the specified id."""
        message_id = self.__build_message_id_object(message_id)
        return self.__stub.getMessage(message_id)

    def get_messages_filters(self) -> ListFilterName:
        """GRPC-API `getMessagesFilters` call returns all the names of sse message filters."""
        return self.__stub.getMessagesFilters(Empty())

    def get_events_filters(self) -> ListFilterName:
        """GRPC-API `getEventsFilters` call returns all the names of sse event filters."""
        return self.__stub.getEventsFilters(Empty())

    def get_event_filter_info(self, filter_name: str) -> FilterInfo:
        """GRPC-API `getEventFilterInfo` call returns event filter info."""
        filter_name = FilterName(filter_name=filter_name)
        return self.__stub.getEventFilterInfo(filter_name)

    def get_message_filter_info(self, filter_name: str) -> FilterInfo:
        """GRPC-API `getMessageFilterInfo` call returns message filter info."""
        filter_name = FilterName(filter_name=filter_name)
        return self.__stub.getMessageFilterInfo(filter_name)

    def match_event(self, event_id: str, filters: List[Filter]) -> IsMatched:
        """GRPC-API `matchEvent` call checks that the event with the specified id is matched by the filter."""
        match_request = MatchRequest(event_id=EventID(id=event_id), filters=filters)
        return self.__stub.matchEvent(match_request)

    def match_message(self, message_id: str, filters: List[Filter]) -> IsMatched:
        """GRPC-API `matchMessage` call checks that the message with the specified id is matched by the filter."""
        message_id = self.__build_message_id_object(message_id)
        match_request = MatchRequest(message_id=message_id, filters=filters)
        return self.__stub.matchMessage(match_request)



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import logging

from th2_data_services.decode_error_handler import UNICODE_REPLACE_HANDLER
from typing import TYPE_CHECKING

from th2_data_services.provider.exceptions import CommandError
from th2_data_services.provider.interfaces import IEventStruct, IMessageStruct, IEventStub, IMessageStub

if TYPE_CHECKING:
    from th2_data_services.provider.v5.interfaces.command import IHTTPProvider5Command

from th2_data_services.provider.interfaces.data_source import IHTTPProviderDataSource
from th2_data_services.provider.v5.struct import (
    provider5_event_struct,
    provider5_message_struct,
    Provider5EventStruct,
    Provider5MessageStruct,
)
from th2_data_services.provider.v5.stub_builder import (
    provider5_event_stub_builder,
    provider5_message_stub_builder,
)
from th2_data_services.provider.v5.provider_api.http import HTTPProvider5API

logger = logging.getLogger(__name__)


class HTTPProvider5DataSource(IHTTPProviderDataSource):
    """DataSource class which provide work with rpt-data-provider.

    Rpt-data-provider version: 5.x.y
    Protocol: HTTP
    """

    def __init__(
        self,
        url: str,
        chunk_length: int = 65536,
        char_enc: str = "utf-8",
        decode_error_handler: str = UNICODE_REPLACE_HANDLER,
        event_struct: IEventStruct = provider5_event_struct,
        message_struct: IMessageStruct = provider5_message_struct,
        event_stub_builder: IEventStub = provider5_event_stub_builder,
        message_stub_builder: IMessageStub = provider5_message_stub_builder,
        check_connect_timeout: (int, float) = 5,
    ):
        """HTTPProvider5DataSource constructor.

        Args:
            url: HTTP data source url.
            check_connect_timeout: How many seconds to wait for the server to send data before giving up.
            chunk_length: How much of the content to read in one chunk.
            char_enc: Encoding for the byte stream.
            decode_error_handler: Registered decode error handler.
            event_struct: Struct of event from rpt-data-provider.
            message_struct: Struct of message from rpt-data-provider.
            event_stub_builder: Stub for event.
            message_stub_builder: Stub for message.
        """
        super().__init__(url, event_struct, message_struct, event_stub_builder, message_stub_builder)

        self._char_enc = char_enc
        self._decode_error_handler = decode_error_handler
        self.__chunk_length = chunk_length
        self.check_connect(check_connect_timeout)
        self._provider_api = HTTPProvider5API(url, chunk_length, decode_error_handler, char_enc)

        logger.info(url)

    def command(self, cmd: IHTTPProvider5Command):
        """HTTP Provider5 command processor.

        Args:
            cmd: The command of data source to execute.

        Returns:
            Data source command result.

        Raises:
            CommandError: If the command was broken.
        """
        try:
            return cmd.handle(data_source=self)
        except Exception as e:
            raise CommandError(f"The command '{cmd.__class__.__name__}' was broken. Details of error:\n{e}")

    @property
    def source_api(self) -> HTTPProvider5API:
        """HTTP Provider5 API."""
        return self._provider_api

    @property
    def event_struct(self) -> Provider5EventStruct:
        """Returns event structure class."""
        return self._event_struct

    @property
    def message_struct(self) -> Provider5MessageStruct:
        """Returns message structure class."""
        return self._message_struct
